"""
Data object for livetiming data
"""

import hashlib
import json
from datetime import timedelta

from fastf1.logger import get_logger
from fastf1.utils import (
    recursive_dict_get,
    to_datetime
)


_logger = get_logger(__name__)


_track_status_mapping = {
    'AllClear': '1',
    'Yellow': '2',
    'SCDeployed': '4',
    'Red': '5',
    'VSCDeployed': '6',
    'VSCEnding': '7'
}


class LiveTimingData:
    """Live timing data object for using saved livetiming data as data source.

    This object is created from data that was recorded using
    :class:`~fastf1.livetiming.client.SignalRClient`. It can be passed to
    various api calling functions using the ``livedata`` keyword.

    Usually you will only instantiate this function and pass it to
    other functions.

    See :mod:`fastf1.livetiming` for a usage example.

    If you want to load data from multiple files you can simply pass multiple
    filenames::

        livedata = LiveTimingData('file1.txt', 'file2.txt', 'file3.txt')

    The files need to be in chronological order but may overlap. I.e. if the
    last five minutes of file1 are the same as the first 5 minutes of file2
    this will be recognized while loading the data. No duplicate data will
    be loaded.

    Args:
        *files (str): One or multiple file names
        remove_duplicates (bool): Remove duplicate lines. Mainly useful when
            loading multiple overlapping recordings. (Checking for duplicates
            is currently very slow for large files. Therefore, it can be
            disabled if this may cause problems.)
    """
    def __init__(self, *files, remove_duplicates=True):
        # file names
        self.files = files
        # parsed data
        self.data = dict()
        # number of json errors
        self.errorcount = 0
        # flag for auto loading on first access
        self._files_read = False
        # date when session was started
        self._start_date = None
        # whether any files were loaded previously, i.e. appending data
        self._previous_files = False
        # hash each line, used to skip duplicates from multiple files
        self._line_hashes = list()
        self._remove_duplicates = remove_duplicates

    def load(self):
        """
        Read all files, parse the data and store it by category.

        Should usually not be called manually. This is called
        automatically the first time :meth:`get`, :meth:`has`
        or :meth:`list_categories` are called.
        """
        _logger.info("Reading live timing data from recording. "
                     "This may take a bit.")
        for fname in self.files:
            self._load_single_file(fname)
        self._files_read = True

    def _load_single_file(self, fname):
        # read one file, parse its content and add it to the already loaded
        # data (if there is data already)
        with open(fname, 'r') as fobj:
            data = fobj.readlines()

        # try to find the correct start date (only if this is the first file)
        if not self._previous_files:
            self._try_set_correct_start_date(data)

        for line in data:
            self._parse_line(line)

        # first file was loaded, others are appended if any more are loaded
        self._previous_files = True

    def _parse_line(self, elem):
        # parse a single line of data

        if self._remove_duplicates:
            # prevent duplicates when loading data (slow, but it works...)
            # allows to load data from overlapping recordings
            lhash = hashlib.md5(elem.encode()).hexdigest()
            if lhash in self._line_hashes:
                return
            self._line_hashes.append(lhash)

        # load the three parts of each data element
        elem = self._fix_json(elem)
        try:
            cat, msg, dt_str = json.loads(elem)
        except (json.JSONDecodeError, ValueError):
            self.errorcount += 1
            return

        # convert string to datetime
        dt = to_datetime(dt_str)
        if dt is None:
            self.errorcount += 1
            return

        # if no start date could be determined beforehand, simply use the
        # first timestamp as we need to have some date as start date;
        # convert timestamp to timedelta (SessionTime) base on start date
        if self._start_date is None:
            self._start_date = dt
            td = timedelta(seconds=0)
        else:
            td = dt - self._start_date

        self._store_message(cat, td, msg)

    def _store_message(self, cat, td, msg):
        # stores parsed messages by category
        # TrackStatus and SessionStatus categories need special handling
        if cat == 'SessionData':
            self._parse_session_data(msg)
        elif cat == 'RaceControlMessages':
            self._parse_race_control_message(msg)
        elif cat not in ('TrackStatus', 'SessionStatus'):
            self._add_to_category(cat, [td, msg])

    def _fix_json(self, elem):
        # fix F1's not json compliant data
        elem = elem.replace("'", '"') \
            .replace('True', 'true') \
            .replace('False', 'false')
        return elem

    def _add_to_category(self, cat, entry):
        if cat not in self.data:
            self.data[cat] = [entry, ]
        else:
            self.data[cat].append(entry)

    def _parse_session_data(self, msg):
        # make sure the categories exist as we want to append to them
        if 'TrackStatus' not in self.data:
            self.data['TrackStatus'] = {'Time': [], 'Status': [],
                                        'Message': []}
        if 'SessionStatus' not in self.data:
            self.data['SessionStatus'] = {'Time': [], 'Status': []}

        if ('StatusSeries' in msg) and isinstance(msg['StatusSeries'], dict):
            for entry in msg['StatusSeries'].values():
                # convert timestamp to timedelta
                try:
                    status_dt = to_datetime(entry['Utc'])
                except (KeyError, ValueError, TypeError):
                    self.errorcount += 1
                    continue
                status_timedelta = status_dt - self._start_date

                # add data to category
                if 'TrackStatus' in entry.keys():
                    status_value = str(entry['TrackStatus'])
                    # convert to numeric system used by the api
                    if not status_value.isnumeric():
                        status_value = _track_status_mapping[status_value]
                    self.data['TrackStatus']['Time'].append(status_timedelta)
                    self.data['TrackStatus']['Status'].append(status_value)
                    self.data['TrackStatus']['Message'].append("")

                elif 'SessionStatus' in entry.keys():
                    self.data['SessionStatus']['Time'].append(status_timedelta)
                    self.data['SessionStatus']['Status'].append(entry['SessionStatus'])

    def _parse_race_control_message(self, msg):
        if 'RaceControlMessages' not in self.data:
            self.data['RaceControlMessages'] = {
                'Utc': [], 'Category': [], 'Message': [], 'Status': [],
                'Flag': [], 'Scope': [], 'Sector': [], 'RacingNumber': []
            }

        if ('Messages' in msg) and isinstance(msg['Messages'], dict):
            for data in msg['Messages'].values():
                for key in ('Utc', 'Category', 'Message', 'Status', 'Flag',
                            'Scope', 'Sector', 'RacingNumber'):
                    self.data['RaceControlMessages'][key].append(data.get(key))

    def _try_set_correct_start_date(self, data):
        # skim content to find 'Started' session status without actually
        # decoding each line to save time
        for elem in data:
            if 'SessionStatus' in elem and 'Started' in elem:
                break
        else:
            # didn't find 'Started'
            _logger.error("Error while trying to set correct "
                          "session start date!")
            return

        # decode matching line
        elem = self._fix_json(elem)
        try:
            cat, msg, dt = json.loads(elem)
        except (json.JSONDecodeError, ValueError):
            _logger.error("Error while trying to set correct "
                          "session start date!")
            return

        # find correct entry in series
        try:
            for entry in msg['StatusSeries']:
                status = recursive_dict_get(entry, 'SessionStatus')
                if status == 'Started':
                    try:
                        self._start_date = to_datetime(entry['Utc'])
                    except (KeyError, ValueError, TypeError):
                        self.errorcount += 1
                        _logger.error("Error while trying to set correct "
                                      "session start date!")
                        return
        except AttributeError:
            for entry in msg['StatusSeries'].values():
                status = entry.get('SessionStatus', None)
                if status == 'Started':
                    try:
                        self._start_date = to_datetime(entry['Utc'])
                    except (KeyError, ValueError, TypeError):
                        self.errorcount += 1
                        _logger.error("Error while trying to set correct "
                                      "session start date!")
                        return

    def get(self, name):
        """
        Return data for category name.

        Will load data on first call, this will take a bit.

        Args:
            name (str): name of the category
            """
        if not self._files_read:
            self.load()
        return self.data[name]

    def has(self, name):
        """
        Check if data for a category name exists.

        Will load data on first call, this will take a bit.

        Args:
            name (str): name of the category
        """
        if not self._files_read:
            self.load()
        return name in self.data.keys()

    def list_categories(self):
        """
        List all available data categories.

        Will load data on first call, this will take a bit.

        Returns:
            list of category names
        """
        if not self._files_read:
            self.load()
        return list(self.data.keys())



"""Driver Laptimes Distribution Visualization
=============================================
Visualizae different drivers' laptime distributions.
"""

import seaborn as sns
from matplotlib import pyplot as plt

import fastf1
import fastf1.plotting


# Enable Matplotlib patches for plotting timedelta values and load
# FastF1's dark color scheme
fastf1.plotting.setup_mpl(mpl_timedelta_support=True, misc_mpl_mods=False,
                          color_scheme='fastf1')


###############################################################################
# Load the race session

race = fastf1.get_session(2023, "Azerbaijan", 'R')
race.load()

###############################################################################
# Get all the laps for the point finishers only.
# Filter out slow laps (yellow flag, VSC, pitstops etc.)
# as they distort the graph axis.
point_finishers = race.drivers[:10]
print(point_finishers)
driver_laps = race.laps.pick_drivers(point_finishers).pick_quicklaps()
driver_laps = driver_laps.reset_index()

###############################################################################
# To plot the drivers by finishing order,
# we need to get their three-letter abbreviations in the finishing order.
finishing_order = [race.get_driver(i)["Abbreviation"] for i in point_finishers]
print(finishing_order)

###############################################################################
# First create the violin plots to show the distributions.
# Then use the swarm plot to show the actual laptimes.

# create the figure
fig, ax = plt.subplots(figsize=(10, 5))

# Seaborn doesn't have proper timedelta support,
# so we have to convert timedelta to float (in seconds)
driver_laps["LapTime(s)"] = driver_laps["LapTime"].dt.total_seconds()

sns.violinplot(data=driver_laps,
               x="Driver",
               y="LapTime(s)",
               hue="Driver",
               inner=None,
               density_norm="area",
               order=finishing_order,
               palette=fastf1.plotting.get_driver_color_mapping(session=race)
               )

sns.swarmplot(data=driver_laps,
              x="Driver",
              y="LapTime(s)",
              order=finishing_order,
              hue="Compound",
              palette=fastf1.plotting.get_compound_mapping(session=race),
              hue_order=["SOFT", "MEDIUM", "HARD"],
              linewidth=0,
              size=4,
              )
# sphinx_gallery_defer_figures

###############################################################################
# Make the plot more aesthetic
ax.set_xlabel("Driver")
ax.set_ylabel("Lap Time (s)")
plt.suptitle("2023 Azerbaijan Grand Prix Lap Time Distributions")
sns.despine(left=True, bottom=True)

plt.tight_layout()
plt.show()



import json
import warnings

from fastf1.ergast.interface import BASE_URL as base_url
from fastf1.ergast.interface import HEADERS as _headers
from fastf1.req import Cache


def fetch_results(year, gp, session):
    """session can be 'Qualifying' or 'Race'
    mainly to port on upper level libraries
    """
    if session == 'Race':
        day = 'results'
        sel = 'Results'
    elif session == 'Qualifying':
        day = 'qualifying'
        sel = 'QualifyingResults'
    elif session in ('Sprint Qualifying', 'Sprint'):
        day = 'sprint'
        sel = 'SprintResults'

    return _parse_ergast(fetch_day(year, gp, day))[0][sel]


def fetch_season(year):
    url = f"{base_url}/{year}.json"
    return _parse_ergast(_parse_json_response(
        Cache.requests_get(url, headers=_headers))
    )


def fetch_day(year, gp, day):
    """day can be 'qualifying' or 'results'
    """
    url = f"{base_url}/{year}/{gp}/{day}.json"
    return _parse_json_response(Cache.requests_get(url, headers=_headers))


def _parse_json_response(r):
    if r.status_code == 200:
        return json.loads(r.content.decode('utf-8'))
    else:
        warnings.warn(f"Request returned: {r.status_code}")
        return None


def _parse_ergast(data):
    return data['MRData']['RaceTable']['Races']



"""Base classes for objects that inherit form Pandas Series or DataFrame."""
from typing import (
    Callable,
    Optional,
    final
)

import pandas as pd


# dangerous import of pandas internals
# these imports are covered by
# test_internals.py::test_pandas_base_internal_imports
# to detect any failures as soon as possible
try:
    from pandas.core.internals import SingleBlockManager
except ImportError as exc:
    _mgr_instance = getattr(pd.Series(dtype=float), '_mgr')
    if _mgr_instance is None:
        raise ImportError("Import of Pandas internals failed. You are likely "
                          "using a recently released version of Pandas that "
                          "isn't yet supported. Please report this issue. In"
                          "the meantime, you can try downgrading to an older "
                          "version of Pandas.") from exc
    SingleBlockManager = type(_mgr_instance)


class BaseDataFrame(pd.DataFrame):
    """Base class for objects that inherit from ``pandas.DataFrame``.

    A DataFrame-like slice of an object that inherits from this class will
    be of equivalent type as the object that it is created from (instead of
    being a Pandas DataFrame).

    A Series-like slice of an object that inherits from this class will by
    default be a normal Pandas Series. An extension to pandas'
    `._constructor_sliced` is implemented in this class to allow using
    different return types for the created Series-like slices of an object
    depending on whether the DataFrame is sliced vertically or horizontally.
    For this, the additional properties ``_constructor_sliced_horizontal``
    and ``_constructor_sliced_vertical`` are introduced. Both properties are
    set to ``pandas.Series`` by default and only need to be overwritten when
    a different type is desired.
    """
    _internal_names = pd.DataFrame._internal_names + ['base_class_view']
    _internal_names_set = set(_internal_names)

    def __repr__(self) -> str:
        return self.base_class_view.__repr__()

    @property
    def _constructor(self) -> Callable[..., "BaseDataFrame"]:
        # by default, use the customized class as a constructor, i.e. all
        # classes that inherit from this base class will always use themselves
        # as a constructor
        return self.__class__

    @final
    @property
    def _constructor_sliced(self) -> Callable[..., pd.Series]:
        # dynamically create a subclass of _BaseSeriesConstructor that
        # has a reference to this self (i.e. the object from which the slice
        # is created) as a class property
        # type(...) returns a new subclass of a Series
        return type('_DynamicBaseSeriesConstructor',  # noqa: return type
                    (_BaseSeriesConstructor,),
                    {'__meta_created_from': self})

    @property
    def _constructor_sliced_horizontal(self) -> Callable[..., pd.Series]:
        return pd.Series

    @property
    def _constructor_sliced_vertical(self) -> Callable[..., pd.Series]:
        return pd.Series

    @property
    def base_class_view(self) -> pd.DataFrame:
        """For a nicer debugging experience; can view DataFrame through
        this property in various IDEs"""
        return pd.DataFrame(self)


class _BaseSeriesConstructor(pd.Series):
    """
    Base class for an intermediary and dynamically defined constructor
    class that implements horizontal and vertical slicing of Pandas DataFrames
    with different result objects types.

    This class is never seen by the user. It is never fully instantiated
    because it always returns an instance of a class that does not derive
    from this class in its __new__ method.
    """

    __meta_created_from: Optional[BaseDataFrame]

    def __new__(cls, data=None, index=None, *args, **kwargs) -> pd.Series:
        parent = getattr(cls, '__meta_created_from', None)

        if ((index is None) and isinstance(data, (pd.Series,
                                                  pd.DataFrame,
                                                  SingleBlockManager))):
            # no index is explicitly given, try to get an index from the
            # data itself
            index = getattr(data, 'index', None)

        if (parent is None) or (index is None):
            # do "conventional" slicing and return a pd.Series
            constructor = pd.Series

        elif parent.index is index:
            # our index matches the parent index, therefore, the data is
            # a column of the parent DataFrame
            constructor = parent._constructor_sliced_vertical
        else:
            # the data is a row of the parent DataFrame
            constructor = parent._constructor_sliced_horizontal

        if (isinstance(data, SingleBlockManager)
                and hasattr(constructor, '_from_mgr')
                and pd.__version__.startswith('2.')):
            obj = constructor._from_mgr(data, axes=data.axes)
        else:
            obj = constructor(data=data, index=index, *args, **kwargs)

        if parent is not None:
            # catch-all fix for some missing __finalize__ calls in Pandas
            # (this is mainly an issue with older versions of Pandas)
            obj.__finalize__(parent)

        return obj


class BaseSeries(pd.Series):
    """Base class for objects that inherit from ``pandas.Series``.

    A same-dimensional slice of an object that inherits from this class will
    be of equivalent type (instead of being a Pandas Series).
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _constructor(self) -> Callable[..., pd.Series]:
        return self.__class__



"""
Tyre strategies during a race
=============================

Plot all drivers' tyre strategies during a race.
"""

from matplotlib import pyplot as plt

import fastf1
import fastf1.plotting


###############################################################################
# Load the race session

session = fastf1.get_session(2022, "Hungary", 'R')
session.load()
laps = session.laps

###############################################################################
# Get the list of driver numbers
drivers = session.drivers
print(drivers)

###############################################################################
# Convert the driver numbers to three letter abbreviations
drivers = [session.get_driver(driver)["Abbreviation"] for driver in drivers]
print(drivers)

###############################################################################
# We need to find the stint length and compound used
# for every stint by every driver.
# We do this by first grouping the laps by the driver,
# the stint number, and the compound.
# And then counting the number of laps in each group.
stints = laps[["Driver", "Stint", "Compound", "LapNumber"]]
stints = stints.groupby(["Driver", "Stint", "Compound"])
stints = stints.count().reset_index()

###############################################################################
# The number in the LapNumber column now stands for the number of observations
# in that group aka the stint length.
stints = stints.rename(columns={"LapNumber": "StintLength"})
print(stints)

###############################################################################
# Now we can plot the strategies for each driver
fig, ax = plt.subplots(figsize=(5, 10))

for driver in drivers:
    driver_stints = stints.loc[stints["Driver"] == driver]

    previous_stint_end = 0
    for idx, row in driver_stints.iterrows():
        # each row contains the compound name and stint length
        # we can use these information to draw horizontal bars
        compound_color = fastf1.plotting.get_compound_color(row["Compound"],
                                                            session=session)
        plt.barh(
            y=driver,
            width=row["StintLength"],
            left=previous_stint_end,
            color=compound_color,
            edgecolor="black",
            fill=True
        )

        previous_stint_end += row["StintLength"]

# sphinx_gallery_defer_figures

###############################################################################
# Make the plot more readable and intuitive
plt.title("2022 Hungarian Grand Prix Strategies")
plt.xlabel("Lap Number")
plt.grid(False)
# invert the y-axis so drivers that finish higher are closer to the top
ax.invert_yaxis()

# sphinx_gallery_defer_figures

###############################################################################
# Plot aesthetics
ax.spines['top'].set_visible(False)
ax.spines['right'].set_visible(False)
ax.spines['left'].set_visible(False)

plt.tight_layout()
plt.show()



from typing import Optional

import requests.exceptions

from fastf1 import __version_short__
from fastf1.mvapi.internals import _logger
from fastf1.req import Cache


PROTO = "https"
HOST = "api.multiviewer.app"
HEADERS = {'User-Agent': f'FastF1/{__version_short__}'}


def _make_url(path: str):
    return f"{PROTO}://{HOST}{path}"


def get_circuit(*, year: int, circuit_key: int) -> Optional[dict]:
    """:meta private:
    Request circuit data from the MultiViewer API and return the JSON
    response."""
    url = _make_url(f"/api/v1/circuits/{circuit_key}/{year}")
    response = Cache.requests_get(url, headers=HEADERS)
    if response.status_code != 200:
        _logger.debug(f"[{response.status_code}] {response.content.decode()}")
        return None

    try:
        return response.json()
    except requests.exceptions.JSONDecodeError:
        return None



"""
.. _GeneralFunctions:

General Functions - :mod:`fastf1`
=================================

.. currentmodule:: fastf1


Accessing Events and Sessions
-----------------------------

When using FastF1, you usually start by loading an event or a
session. This can be done with one of the following functions:

.. autosummary::
    fastf1.get_session
    fastf1.get_testing_session
    fastf1.get_event
    fastf1.get_events_remaining
    fastf1.get_testing_session
    fastf1.get_event_schedule


.. _requests-and-caching:

Requests and Caching
--------------------

.. automodule::
    fastf1.req

.. currentmodule:: fastf1


General Functions - API Reference
---------------------------------

Event and Session API
.....................

.. autofunction:: get_session
.. autofunction:: get_testing_session
.. autofunction:: get_event
.. autofunction:: get_events_remaining
.. autofunction:: get_testing_event
.. autofunction:: get_event_schedule

Cache API
.........

.. autoclass:: Cache
    :members:
        enable_cache,
        clear_cache,
        get_cache_info,
        disabled,
        set_disabled,
        set_enabled,
        offline_mode
    :autosummary:


.. autoclass:: RateLimitExceededError


Configure Logging Verbosity
...........................

All parts of FastF1 generally log at the log level 'INFO'.
The reason for this is that many data loading processes take multiple
seconds to complete. Logging is used to give progress information here as well
as for showing warnings and non-terminal errors.

The logging level for FastF1 can be easily customized::

    import fastf1

    fastf1.set_log_level('WARNING')

    # ... your code  here ... #

The available levels are (in order of increasing severity): DEBUG, INFO,
WARNING, ERROR and CRITICAL.

.. autofunction:: set_log_level

For more information see :ref:`logging`.

"""
try:
    from . import _version
except ImportError:
    _version = None

__version__ = getattr(_version, 'version', '0.0+UNKNOWN')
__version_tuple__ = getattr(_version, 'version_tuple', (0, 0, '+UNKNOWN'))
if __version_tuple__:
    # create a short version containing only the public version
    __version_short__ = ".".join(str(digit) for digit in __version_tuple__
                                 if str(digit).isnumeric())
else:
    __version_short__ = __version__


from typing import Dict

from fastf1.events import get_session  # noqa: F401
from fastf1.events import (  # noqa: F401
    get_event,
    get_event_schedule,
    get_events_remaining,
    get_testing_event,
    get_testing_session
)
from fastf1.logger import set_log_level  # noqa: F401
from fastf1.req import (  # noqa: F401
    Cache,
    RateLimitExceededError
)


_DRIVER_TEAM_MAPPING: Dict[str, Dict[str, str]] = {
    # only necessary when loading live timing data that does not include
    # the driver and team listing and no data is available on ergast yet
    '23': {'Abbreviation': 'ALB', 'FirstName': 'Alexander',
           'LastName': 'Albon', 'TeamName': 'Williams'},
    '14': {'Abbreviation': 'ALO', 'FirstName': 'Fernando',
           'LastName': 'Alonso', 'TeamName': 'Alpine F1 Team'},
    '77': {'Abbreviation': 'BOT', 'FirstName': 'Valtteri',
           'LastName': 'Bottas', 'TeamName': 'Alfa Romeo'},
    '10': {'Abbreviation': 'GAS', 'FirstName': 'Pierre',
           'LastName': 'Gasly', 'TeamName': 'AlphaTauri'},
    '44': {'Abbreviation': 'HAM', 'FirstName': 'Lewis',
           'LastName': 'Hamilton', 'TeamName': 'Mercedes'},
    '27': {'Abbreviation': 'HUL', 'FirstName': 'Nico',
           'LastName': 'Hülkenberg', 'TeamName': 'Aston Martin'},
    '6': {'Abbreviation': 'LAT', 'FirstName': 'Nicholas',
          'LastName': 'Latifi', 'TeamName': 'Williams'},
    '16': {'Abbreviation': 'LEC', 'FirstName': 'Charles',
           'LastName': 'Leclerc', 'TeamName': 'Ferrari'},
    '20': {'Abbreviation': 'MAG', 'FirstName': 'Kevin',
           'LastName': 'Magnussen', 'TeamName': 'Haas F1 Team'},
    '4': {'Abbreviation': 'NOR', 'FirstName': 'Lando',
          'LastName': 'Norris', 'TeamName': 'McLaren'},
    '31': {'Abbreviation': 'OCO', 'FirstName': 'Esteban',
           'LastName': 'Ocon', 'TeamName': 'Alpine F1 Team'},
    '11': {'Abbreviation': 'PER', 'FirstName': 'Sergio',
           'LastName': 'Pérez', 'TeamName': 'Red Bull'},
    '3': {'Abbreviation': 'RIC', 'FirstName': 'Daniel',
          'LastName': 'Ricciardo', 'TeamName': 'McLaren'},
    '63': {'Abbreviation': 'RUS', 'FirstName': 'George',
           'LastName': 'Russell', 'TeamName': 'Mercedes'},
    '55': {'Abbreviation': 'SAI', 'FirstName': 'Carlos',
           'LastName': 'Sainz', 'TeamName': 'Ferrari'},
    '47': {'Abbreviation': 'MSC', 'FirstName': 'Mick',
           'LastName': 'Schumacher', 'TeamName': 'Haas F1 Team'},
    '18': {'Abbreviation': 'STR', 'FirstName': 'Lance',
           'LastName': 'Stroll', 'TeamName': 'Aston Martin'},
    '22': {'Abbreviation': 'TSU', 'FirstName': 'Yuki',
           'LastName': 'Tsunoda', 'TeamName': 'AlphaTauri'},
    '1': {'Abbreviation': 'VER', 'FirstName': 'Max',
          'LastName': 'Verstappen', 'TeamName': 'Red Bull'},
    '5': {'Abbreviation': 'VET', 'FirstName': 'Sebastian',
          'LastName': 'Vettel', 'TeamName': 'Aston Martin'},
    '24': {'Abbreviation': 'ZHO', 'FirstName': 'Guanyu',
           'LastName': 'Zhou', 'TeamName': 'Alfa Romeo'}
}



# imports for exposed names
from fastf1.ergast.interface import Ergast  # noqa: F401
from fastf1.ergast.legacy import fetch_results  # noqa: F401
from fastf1.ergast.legacy import (  # noqa: F401
    fetch_day,
    fetch_season
)



"""Using the Fast-F1 signalr client?
======================================

Demonstrates the usage of the SignalRClient
"""
import logging

from fastf1.livetiming.client import SignalRClient


log = logging.getLogger()
log.setLevel(logging.DEBUG)

client = SignalRClient(filename="output.txt", debug=True)
client.start()



import argparse
import sys

from fastf1.livetiming.client import (
    SignalRClient,
    messages_from_raw
)


def save(args):
    mode = 'a' if args.append else 'w'
    client = SignalRClient(args.file, filemode=mode, debug=args.debug,
                           timeout=args.timeout)
    client.start()


def convert(args):
    with open(args.input, 'r') as infile:
        messages = infile.readlines()
    data, ec = messages_from_raw(messages)
    with open(args.output, 'w') as outfile:
        for elem in data:
            outfile.write(str(elem)+'\n')
    print(f"Completed with {ec} error(s)")


parser = argparse.ArgumentParser(
    prog="python -m fastf1.livetiming",
    description="Save live timing data during a session",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter
)

subparsers = parser.add_subparsers()

rec_parser = subparsers.add_parser(
    'save', help='Save live timing data'
)
conv_parser = subparsers.add_parser(
    'extract', help='Extract messages from saved debug-mode data'
)

rec_parser.add_argument('file', type=str, help='Output file name')
rec_parser.add_argument('--append', action='store_true', default=False,
                        help="Append to output file. By default the file is "
                             "overwritten if it exists already.")
rec_parser.add_argument('--debug', action='store_true', default=False,
                        help='Enable debug mode: save full SignalR message, '
                             'not just the data.')
rec_parser.add_argument('--timeout', type=int, default=60,
                        help='Timeout in seconds after which the client will '
                             'automatically exit if no data is received.')
rec_parser.set_defaults(func=save)

conv_parser.add_argument("input", type=str, help='Input file name')
conv_parser.add_argument("output", type=str, help='Output file name')
conv_parser.set_defaults(func=convert)

if not len(sys.argv) > 1:
    # user did not provide any arguments
    parser.print_help()
    parser.exit(1)

args = parser.parse_args()
args.func(args)  # call function associated with subparser



"""Draw a track map with numbered corners
=========================================

Use the position data of a single lap to draw a track map.
Then annotate the map with corner numbers.
"""
##############################################################################
# Import FastF1 and load the data. Use the telemetry from the fastest for the
# track map. (You could also use any other lap instead.)

import matplotlib.pyplot as plt
import numpy as np

import fastf1


session = fastf1.get_session(2023, 'Silverstone', 'Q')
session.load()

lap = session.laps.pick_fastest()
pos = lap.get_pos_data()

circuit_info = session.get_circuit_info()


##############################################################################
# Define a helper function for rotating points around the origin of the
# coordinate system.
#
# The matrix ``[[cos, sin], [-sin, cos]]`` is called a rotation matrix.
#
# By matrix multiplication of the rotation matrix with a vector [x, y], a new
# rotated vector [x_rot, y_rot] is obtained.
# (See also: https://en.wikipedia.org/wiki/Rotation_matrix)

def rotate(xy, *, angle):
    rot_mat = np.array([[np.cos(angle), np.sin(angle)],
                        [-np.sin(angle), np.cos(angle)]])
    return np.matmul(xy, rot_mat)


##############################################################################
# Get the coordinates of the track map from the telemetry of the lap and
# rotate the coordinates using the rotation from ``circuit_info`` so that
# the track map is oriented correctly. After that, plot the rotated track map.

# Get an array of shape [n, 2] where n is the number of points and the second
# axis is x and y.
track = pos.loc[:, ('X', 'Y')].to_numpy()

# Convert the rotation angle from degrees to radian.
track_angle = circuit_info.rotation / 180 * np.pi

# Rotate and plot the track map.
rotated_track = rotate(track, angle=track_angle)
plt.plot(rotated_track[:, 0], rotated_track[:, 1])

# sphinx_gallery_defer_figures


##############################################################################
# Finally, the corner markers are plotted. To plot the numbers next to the
# track, an offset vector that points straight up is defined. This offset
# vector is then rotated by the angle that is given for each corner marker.
# A line and circular bubble are drawn and the corner marker text is printed
# inside the bubble.

offset_vector = [500, 0]  # offset length is chosen arbitrarily to 'look good'

# Iterate over all corners.
for _, corner in circuit_info.corners.iterrows():
    # Create a string from corner number and letter
    txt = f"{corner['Number']}{corner['Letter']}"

    # Convert the angle from degrees to radian.
    offset_angle = corner['Angle'] / 180 * np.pi

    # Rotate the offset vector so that it points sideways from the track.
    offset_x, offset_y = rotate(offset_vector, angle=offset_angle)

    # Add the offset to the position of the corner
    text_x = corner['X'] + offset_x
    text_y = corner['Y'] + offset_y

    # Rotate the text position equivalently to the rest of the track map
    text_x, text_y = rotate([text_x, text_y], angle=track_angle)

    # Rotate the center of the corner equivalently to the rest of the track map
    track_x, track_y = rotate([corner['X'], corner['Y']], angle=track_angle)

    # Draw a circle next to the track.
    plt.scatter(text_x, text_y, color='grey', s=140)

    # Draw a line from the track to this circle.
    plt.plot([track_x, text_x], [track_y, text_y], color='grey')

    # Finally, print the corner number inside the circle.
    plt.text(text_x, text_y, txt,
             va='center_baseline', ha='center', size='small', color='white')

# sphinx_gallery_defer_figures


##############################################################################
# Add a title, remove tick labels to clean up the plot, set equal axis ratio,
# so that the track is not distorted and show the plot.

plt.title(session.event['Location'])
plt.xticks([])
plt.yticks([])
plt.axis('equal')
plt.show()



import json
from typing import get_type_hints

import pandas as pd
from docutils import nodes
from docutils.parsers.rst import (
    Directive,
    directives
)

import fastf1.ergast.structure
from fastf1 import __version__
from fastf1.ergast.interface import ErgastResultFrame


class ApiMappingDirective(Directive):
    """A Sphinx directive to generate a description of an Ergast API endpoint
    from the object-oriented structure description in
    :mod:`fastf1.ergast.structure`.

    Required Arguments:

        - category (str): name of the category object in ``.structure`` that
          should be described

    Optional Arguments:

        - ``:subcategory: <name>``: name of the subcategory object that is
          split out from the response during flattening of the data
          (MultiResponse)

        - ``:describe-dataframe:``: flat that indicates that a description of
          the flattened data frame should be generated as well
    """
    required_arguments = 1
    option_spec = {
        'subcategory': directives.unchanged,
        'describe-dataframe': directives.flag
    }

    def _build_tree(self, category, excluded_category, *,
                    include_new_name=True):
        # Creates an object that describes the specified 'category' that
        # matches the structure of the corresponding response from ergast.
        # All values within this structure are strings that specify the data
        # type of the value in the real response and optionally also specify
        # the new name of the key after renaming (if 'include_new_name' is
        # True). If 'excluded_category' is not None, the subkey within the
        # response that matches the name of the excluded category is skipped
        # when building the object. This is used to create split
        # representations for description and content of MultiResponse objects.
        result = dict()

        for name, mapping in category['map'].items():
            # get the nicely readable name of the data type after auto casting
            _type = mapping['type']
            type_hints = get_type_hints(_type).get('return')
            if type_hints:
                # if casting is done through a conversion function like
                # `ergast.structure.time_from_ergast`, the resulting data
                # type needs to be determined from type hints
                type_str = getattr(type_hints, '__name__', str(type_hints))
            else:
                # common data types like int, float, ...: use .__name__
                type_str = _type.__name__

            if include_new_name:
                result[name] = f"{mapping['name']} <{type_str}>"
            else:
                result[name] = f"<{type_str}>"

        # recursively process subkeys
        for subcategory in category['sub']:
            if subcategory is excluded_category:
                continue

            subcontent = self._build_tree(subcategory, excluded_category,
                                          include_new_name=include_new_name)

            result[subcategory['name']] = subcontent

        # change to correct representation for keys that contain a list
        # of items
        if category['type'] is list:
            result = [result]

        return result

    def run(self):
        # get the category and splitcategory objects by name
        # splitcategory only exists for API endpoints that return
        # ErgastMultiResponse objects. It defines the subkey that is split
        # from the description df into the content df.
        root_category = getattr(fastf1.ergast.structure, self.arguments[0])
        splitcategory_name = self.options.get('subcategory')
        if splitcategory_name:
            splitcategory = getattr(fastf1.ergast.structure,
                                    splitcategory_name)
        else:
            splitcategory = None

        # First, build a full tree of the response to show the structure of
        # the original response together with the new names of keys and the
        # new data types of their values after auto casting.
        tree = self._build_tree(root_category, None)
        result_str = json.dumps(tree, indent=4).replace('"', '')

        nodes_list = \
            [nodes.raw("",
                       "<details><summary><a>API Mapping</a></summary>"
                       "Structure of the raw response, renamed key names for "
                       "flattening and dtypes for automatic type casting:"
                       "</br></br>",
                       format='html'),
             nodes.literal_block("", result_str, language='none'),
             nodes.raw("", "</details>", format='html')]

        if 'describe-dataframe' in self.options:
            # Second, add a description of how the data looks after flattening
            # in the pandas responses

            nodes_list.append(nodes.raw(
                "",
                "<details><summary><a>DataFrame Description</a></summary>"
                "DataFrame column names and dtypes for automatic type"
                "casting:</br></br>",
                format='html'
            ))

            result_str = self._generate_df_description(root_category,
                                                       splitcategory)

            if splitcategory:
                # add subheading if the response is split into a MultiResponse
                nodes_list.append(nodes.raw(
                    "",
                    "<code>ErgastMultiResponse.description</code>",
                    format='html'
                ))

            nodes_list.append(nodes.literal_block(
                "", result_str, language='none'
            ))

            if splitcategory:
                # for split responses, generate the second description of the
                # previously excluded subkey
                nodes_list.append(nodes.raw(
                    "",
                    f"<code>ErgastMultiResponse.content</code> (contains data "
                    f"from subkey <code>{splitcategory_name}</code>)",
                    format='html'
                ))

                result_str = self._generate_df_description(splitcategory, None)
                nodes_list.append(nodes.literal_block(
                    "", result_str, language='none'
                ))

            nodes_list.append(nodes.raw("", "</details>", format='html'))

        return nodes_list

    def _generate_df_description(self, root_category, splitcategory):
        # generate an object that describes the response in the same
        # structure as the response itself
        tree = self._build_tree(root_category, splitcategory,
                                include_new_name=False)
        # process this object the same way an actual response is processed, so
        # that the same renaming and transformations of the structure are
        # applied to it
        flat = ErgastResultFrame._prepare_response(tree, root_category,
                                                   cast=False)
        # print the flattened data nicely formatted as a table (use str
        # representation of a dataframe for this)
        # remove the last line (unnecessary info about dataframe)
        result_str = '\n'.join(
            str(pd.DataFrame(flat).iloc[0]).split('\n')[:-1]
        )
        return result_str


def setup(app):
    """ Register directive with Sphinx """
    app.add_directive('ergast-api-map', ApiMappingDirective)
    return {'version': __version__}



from typing import List

import numpy as np
from pandas import (
    DataFrame,
    Index,
    RangeIndex
)

from fastf1.internals import internals_logger as logger


try:
    # import internal pandas functions and objects
    from pandas.core.internals.construction import (
        BlockPlacement,
        _get_axes,
        create_block_manager_from_blocks,
        new_block_2d
    )
    from pandas.core.internals.managers import _consolidate

    # verify existence of non-public methods of public objects
    if not hasattr(Index, '_with_infer'):
        raise AttributeError

except (AttributeError, ImportError) as import_exc:
    logger.warning("Import of pandas internals failed. Performance will be "
                   "reduced.")
    logger.debug("Traceback for import failure", exc_info=import_exc)
    CREATE_DF_FAST = False
else:
    CREATE_DF_FAST = True


def create_df_fast(
        *,
        arrays: List[np.ndarray],
        columns: list,
        fallback: bool = True
) -> DataFrame:
    """Implements fast creation of DataFrames.

    This function tries to speed up DataFrame creation by skipping many
    pandas internal steps that are not required. This only works for some
    data and only if this data is correctly processed beforehand.

    In case of error, this function falls back to the official documented
    way of creating DataFrames.

    Args:
        arrays: list of 1D numpy arrays of equal length
        columns: list of column names (one name per array)
        fallback: use Pandas' default method of DataFrame creation in case
            of errors

    Returns:
        Pandas DataFrame equivalent to the one create with the default
        dataframe constructor

    """
    try:
        return _unsafe_create_df_fast(arrays, columns)
    except Exception as exc:
        if not fallback:
            raise exc
        # in case of error, use the usual but slower method
        logger.warning("Falling back to slow data frame creation!")
        logger.debug("Error during fast DataFrame creation", exc_info=exc)
        return _fallback_create_df(arrays, columns)


def _fallback_create_df(
        arrays: List[np.ndarray],
        columns: list
) -> DataFrame:
    data = {col: arr for col, arr in zip(columns, arrays)}
    return DataFrame(data)


def _fallback_if_unsupported(func):
    if not CREATE_DF_FAST:
        return _fallback_create_df
    else:
        return func


@_fallback_if_unsupported
def _unsafe_create_df_fast(
        arrays: List[np.ndarray],
        columns: list
) -> DataFrame:
    # Implements parts of pandas' internal DataFrame creation mechanics
    # but skipping unnecessary tests and most importantly sanitization and
    # validation of the data. This results in much higher performance but is
    # not necessarily safe, for obvious reasons.
    # This method should only be used in FastF1 internal code where, type
    # casting, array creation and data sanitization is very explicitly already
    # performed. Always use this method through `create_df_fast`, so that there
    # is a fallback to safe DataFrame creation in case of an error.
    index = RangeIndex(0, len(arrays[0]))
    columns = Index._with_infer(
        list(columns), copy=False, tupleize_cols=False
    )

    index, columns = _get_axes(
        len(index), len(columns), index=index, columns=columns
    )

    block_values = list()
    for n, arr in enumerate(arrays):
        values = arr.reshape(-1, 1).T
        nb = new_block_2d(values, placement=BlockPlacement(n))
        block_values.append(nb)

    block_values = list(_consolidate(tuple(block_values)))

    mgr = create_block_manager_from_blocks(
            block_values, [columns, index], verify_integrity=False
    )

    df = DataFrame(mgr)

    return df



import asyncio
import concurrent.futures
import json
import logging
import time
from typing import (
    Iterable,
    Optional
)

import requests

import fastf1
from fastf1.signalr_aio import Connection


def messages_from_raw(r: Iterable):
    """Extract data messages from raw recorded SignalR data.

    This function can be used to extract message data from raw SignalR data
    which was saved using :class:`SignalRClient` in debug mode.

    Args:
        r: Iterable containing raw SignalR responses.
    """
    ret = list()
    errorcount = 0
    for data in r:
        # fix F1's not json compliant data
        data = data.replace("'", '"') \
            .replace('True', 'true') \
            .replace('False', 'false')
        try:
            data = json.loads(data)
        except json.JSONDecodeError:
            errorcount += 1
            continue
        messages = data['M'] if 'M' in data and len(data['M']) > 0 else {}
        for inner_data in messages:
            hub = inner_data['H'] if 'H' in inner_data else ''
            if hub.lower() == 'streaming':
                # method = inner_data['M']
                message = inner_data['A']
                ret.append(message)

    return ret, errorcount


class SignalRClient:
    """A client for receiving and saving F1 timing data which is streamed
    live over the SignalR protocol.

    During an F1 session, timing data and telemetry data are streamed live
    using the SignalR protocol. This class can be used to connect to the
    stream and save the received data into a file.

    The data will be saved in a raw text format without any postprocessing.
    It is **not** possible to use this data during a session. Instead, the
    data can be processed after the session using the :mod:`fastf1.api` and
    :mod:`fastf1.core`

    Args:
        filename: filename (opt. with path) for the output file
        filemode: one of 'w' or 'a'; append to or overwrite
            file content it the file already exists. Append-mode may be useful
            if the client is restarted during a session.
        debug: When set to true, the complete SignalR
            message is saved. By default, only the actual data from a
            message is saved.
        timeout: Number of seconds after which the client
            will automatically exit when no message data is received.
            Set to zero to disable.
        logger: By default, errors are logged to the console. If you wish to
            customize logging, you can pass an instance of
            :class:`logging.Logger` (see: :mod:`logging`).
    """
    _connection_url = 'https://livetiming.formula1.com/signalr'

    def __init__(self, filename: str, filemode: str = 'w', debug: bool = False,
                 timeout: int = 60, logger: Optional = None):

        self.headers = {'User-agent': 'BestHTTP',
                        'Accept-Encoding': 'gzip, identity',
                        'Connection': 'keep-alive, Upgrade'}

        self.topics = ["Heartbeat", "CarData.z", "Position.z",
                       "ExtrapolatedClock", "TopThree", "RcmSeries",
                       "TimingStats", "TimingAppData",
                       "WeatherData", "TrackStatus", "DriverList",
                       "RaceControlMessages", "SessionInfo",
                       "SessionData", "LapCount", "TimingData"]

        self.debug = debug
        self.filename = filename
        self.filemode = filemode
        self.timeout = timeout
        self._connection = None

        if not logger:
            logging.basicConfig(
                format="%(asctime)s - %(levelname)s: %(message)s"
            )
            self.logger = logging.getLogger('SignalR')
        else:
            self.logger = logger

        self._output_file = None
        self._t_last_message = None

    def _to_file(self, msg):
        self._output_file.write(msg + '\n')
        self._output_file.flush()

    async def _on_do_nothing(self, msg):
        # just do nothing with the message; intended for debug mode where some
        # callback method still needs to be provided
        pass

    async def _on_message(self, msg):
        self._t_last_message = time.time()
        loop = asyncio.get_running_loop()
        try:
            with concurrent.futures.ThreadPoolExecutor() as pool:
                await loop.run_in_executor(
                    pool, self._to_file, str(msg)
                )
        except Exception:
            self.logger.exception("Exception while writing message to file")

    async def _on_debug(self, **data):
        if 'M' in data and len(data['M']) > 0:
            self._t_last_message = time.time()

        loop = asyncio.get_running_loop()
        try:
            with concurrent.futures.ThreadPoolExecutor() as pool:
                await loop.run_in_executor(
                    pool, self._to_file, str(data)
                )
        except Exception:
            self.logger.exception("Exception while writing message to file")

    async def _run(self):
        self._output_file = open(self.filename, self.filemode)
        # Create connection
        session = requests.Session()
        session.headers = self.headers
        self._connection = Connection(self._connection_url, session=session)

        # Register hub
        hub = self._connection.register_hub('Streaming')

        if self.debug:
            # Assign error handler
            self._connection.error += self._on_debug
            # Assign debug message handler to save raw responses
            self._connection.received += self._on_debug
            hub.client.on('feed', self._on_do_nothing)
            # need to connect an async method
        else:
            # Assign hub message handler
            hub.client.on('feed', self._on_message)

        hub.server.invoke("Subscribe", self.topics)

        # Start the client
        loop = asyncio.get_event_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            await loop.run_in_executor(pool, self._connection.start)

    async def _supervise(self):
        self._t_last_message = time.time()
        while True:
            if (self.timeout != 0
                    and time.time() - self._t_last_message > self.timeout):
                self.logger.warning(f"Timeout - received no data for more "
                                    f"than {self.timeout} seconds!")
                self._connection.close()
                return
            await asyncio.sleep(1)

    async def _async_start(self):
        self.logger.info(f"Starting FastF1 live timing client "
                         f"[v{fastf1.__version__}]")
        await asyncio.gather(asyncio.ensure_future(self._supervise()),
                             asyncio.ensure_future(self._run()))
        self._output_file.close()
        self.logger.warning("Exiting...")

    def start(self):
        """Connect to the data stream and start writing the data to a file."""
        try:
            asyncio.run(self._async_start())
        except KeyboardInterrupt:
            self.logger.warning("Keyboard interrupt - exiting...")
            return



"""Speed visualization on track map
======================================

(Example provided by @JSEHV on Github)
"""
import matplotlib as mpl
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.collections import LineCollection

import fastf1 as ff1


##############################################################################
# First, we define some variables that allow us to conveniently control what
# we want to plot.
year = 2021
wknd = 9
ses = 'R'
driver = 'RIC'
colormap = mpl.cm.plasma


##############################################################################
# Next, we load the session and select the desired data.
session = ff1.get_session(year, wknd, ses)
weekend = session.event
session.load()
lap = session.laps.pick_driver(driver).pick_fastest()

# Get telemetry data
x = lap.telemetry['X']              # values for x-axis
y = lap.telemetry['Y']              # values for y-axis
color = lap.telemetry['Speed']      # value to base color gradient on


##############################################################################
# Now, we create a set of line segments so that we can color them
# individually. This creates the points as a N x 1 x 2 array so that we can
# stack points  together easily to get the segments. The segments array for
# line collection needs to be (numlines) x (points per line) x 2 (for x and y)
points = np.array([x, y]).T.reshape(-1, 1, 2)
segments = np.concatenate([points[:-1], points[1:]], axis=1)


##############################################################################
# After this, we can actually plot the data.

# We create a plot with title and adjust some setting to make it look good.
fig, ax = plt.subplots(sharex=True, sharey=True, figsize=(12, 6.75))
fig.suptitle(f'{weekend.name} {year} - {driver} - Speed', size=24, y=0.97)

# Adjust margins and turn of axis
plt.subplots_adjust(left=0.1, right=0.9, top=0.9, bottom=0.12)
ax.axis('off')


# After this, we plot the data itself.
# Create background track line
ax.plot(lap.telemetry['X'], lap.telemetry['Y'],
        color='black', linestyle='-', linewidth=16, zorder=0)

# Create a continuous norm to map from data points to colors
norm = plt.Normalize(color.min(), color.max())
lc = LineCollection(segments, cmap=colormap, norm=norm,
                    linestyle='-', linewidth=5)

# Set the values used for colormapping
lc.set_array(color)

# Merge all line segments together
line = ax.add_collection(lc)


# Finally, we create a color bar as a legend.
cbaxes = fig.add_axes([0.25, 0.05, 0.5, 0.05])
normlegend = mpl.colors.Normalize(vmin=color.min(), vmax=color.max())
legend = mpl.colorbar.ColorbarBase(cbaxes, norm=normlegend, cmap=colormap,
                                   orientation="horizontal")


# Show the plot
plt.show()



from typing import Dict

from fastf1.plotting._constants.base import (
    Compounds,
    Team,
    TeamColors
)


Teams: Dict[str, Team] = {
    'ferrari': Team(
        ShortName='Ferrari',
        TeamColor=TeamColors(
            Official='#dc0000',
            Default='#dc0000'
        )
    ),
    'force india': Team(
        ShortName='Force India',
        TeamColor=TeamColors(
            Official='#f596c8',
            Default='#ff87bc'
        )
    ),
    'haas': Team(
        ShortName='Haas',
        TeamColor=TeamColors(
            Official='#828282',
            Default='#b6babd'
        )
    ),
    'mclaren': Team(
        ShortName='McLaren',
        TeamColor=TeamColors(
            Official='#ff8000',
            Default='#ff8000'
        )
    ),
    'mercedes': Team(
        ShortName='Mercedes',
        TeamColor=TeamColors(
            Official='#00d2be',
            Default='#00f5d0'
        )
    ),
    'racing point': Team(
        ShortName='Racing Point',
        TeamColor=TeamColors(
            Official='#f596c8',
            Default='#ff87bc'
        )
    ),
    'red bull': Team(
        ShortName='Red Bull',
        TeamColor=TeamColors(
            Official='#1e41ff',
            Default='#1e41ff'
        )
    ),
    'renault': Team(
        ShortName='Renault',
        TeamColor=TeamColors(
            Official='#fff500',
            Default='#fff500'
        )
    ),
    'sauber': Team(
        ShortName='Sauber',
        TeamColor=TeamColors(
            Official='#9b0000',
            Default='#900000'
        )
    ),
    'toro rosso': Team(
        ShortName='Toro Rosso',
        TeamColor=TeamColors(
            Official='#469bff',
            Default='#2b4562'
        )
    ),
    'williams': Team(
        ShortName='Williams',
        TeamColor=TeamColors(
            Official='#ffffff',
            Default='#00a0dd'
        )
    )
}

CompoundColors: Dict[Compounds, str] = {
    Compounds.HyperSoft: "#feb1c1",
    Compounds.UltraSoft: "#b24ba7",
    Compounds.SuperSoft: "#fc2b2a",
    Compounds.Soft: "#ffd318",
    Compounds.Medium: "#f0f0f0",
    Compounds.Hard: "#00a2f5",
    Compounds.SuperHard: "#fd7d3c",
    Compounds.Intermediate: "#43b02a",
    Compounds.Wet: "#0067ad",
    Compounds.Unknown: "#00ffff",
    Compounds.TestUnknown: "#434649"
}



"""
.. warning::
    :mod:`fastf1.api` will be considered private in future releases and
    potentially be removed or changed. Please do not use functionality from
    :mod:`fastf1.api`. If you currently require functionality from there,
    please open an issue on Github with details about what you require and why.

A collection of functions to interface with the F1 web api.

.. autosummary::
   :nosignatures:

   timing_data
   timing_app_data
   car_data
   position_data
   track_status_data
   session_status_data
   race_control_messages
   lap_count
   driver_info
   weather_data
   fetch_page
   parse

"""
import warnings

from fastf1._api import *  # noqa


warnings.warn("`fastf1.api` will be considered private in future releases and "
              "potentially be removed or changed!", UserWarning)



"""
Live Timing Client - :mod:`fastf1.livetiming`
=============================================

.. warning::
    The live timing client does only support Python 3.8 or 3.9!

This module can be used to save live timing data during a session.
It is **not** possible to do real-time processing of the data.


There are two ways to interact with this module.

    - From a python script by creating an instance of
      :class:`.SignalRClient`

    - From the command line by calling ``python -m fastf1.livetiming``


Usage Example
-------------

1. Record live timing data during a session

.. code-block:: console

    python -m fastf1.livetiming save saved_data.txt


2. Load this data into FastF1 after the session has finished

.. code-block::

    import fastf1
    from fastf1.livetiming.data import LiveTimingData

    livedata = LiveTimingData('saved_data.txt')
    session = fastf1.get_testing_session(2021, 1, 1)
    session.load(livedata=livedata)

optionally you can load live timing data from two or more files

.. code-block::

    livedata = LiveTimingData('saved_data_1.txt', 'saved_data_2.txt')




Important Notes
---------------

- You should try to record a full session whenever possible. Recording should
  start 2-3 minutes before a sessions starts. The api parser may not be able to
  deal with the data correctly if the beginning of a session is missing.
  Your mileage may vary depending on what and how much data is missing.

- You should not mix recorded live timing data and data requested from the api
  after a session. The data will likely not be synchronized correctly.

- You should use the cache with saved live timing data too. This will speed up
  loading of the data considerably after the first run.

- You need to use different cache directories if you want to cache
  saved live timing data and api data **for the same session**!
  The cache cannot tell the data sources apart. If cached data from one source
  already exists it will not be reloaded automatically from a different source.
  See :class:`fastf1.req.Cache` for details on how to configure a custom
  cache directory. One simple way is to add the following line to your code::

    fastf1.Cache.enable_cache('path/to/cache/directory')

- You need to force a cache update if you modify the
  :class:`~fastf1.livetiming.data.LiveTimingData` (e.g. by adding a second
  file). The cache cannot tell that the input source was modified and will
  still use the old cached data.

  You can force-refresh the cache::

    fastf1.Cache.enable_cache('path/to/cache/directory', force_renew=True)

  You only need to use ``forece_renew=True`` once after modifying the
  input data.

- The SignalR Client seems to get disconnected after 2 hours of recording. It
  looks like the connection is terminated by the server. You need to manually
  start a second recording before the first one disconnects if you want to
  have no gap in your recording.

  Use a different output file name for the second (or any subsequent)
  recording. You can then load :class:`.data.LiveTimingData` from multiple
  files. The files need to be provided in chronological order. The content of
  the files may overlap. Data from overlapping recordings is recognized and
  will not be loaded as a duplicate.




Command Line Interface
----------------------

Available commands and options when calling ``python -m fastf1.livetiming``

The module has two main commands

.. code-block:: console

      {save,extract}
        save          Save live timing data
        extract       Extract messages from saved debug-mode data

Save
++++

**The main command for recording live timing data during a session**

.. code-block:: console

    usage: python -m fastf1.livetiming save [-h] [--append] [--debug]
           [--timeout TIMEOUT] file

    positional arguments:
      file               Output file name

    optional arguments:
      -h, --help         show this help message and exit
      --append           Append to output file. By default the file is
                         overwritten if it exists already.
      --debug            Enable debug mode: save full SignalR message, not
                         just the data.
      --timeout TIMEOUT  Timeout in seconds after which the client will
                         automatically exit if no data is received


Extract
+++++++

**Only for when data was saved with the optional '--debug' argument**

Recording in debug mode saves the full SignalR messages as received. The non
debug mode saves only the important data part of a message. The data part of
each message needs to be extracted to utilize the debug-mode data.
The extracted data is the same data you get when saving without the '--debug'
argument.

.. code-block:: console

    usage: python -m fastf1.livetiming extract [-h] input output

    positional arguments:
      input       Input file name
      output      Output file name

    optional arguments:
      -h, --help  show this help message and exit




Live Timing Client
------------------

.. automodule:: fastf1.livetiming.client
    :members:
    :undoc-members:
    :show-inheritance:


Live Timing Data Object
-----------------------

.. automodule:: fastf1.livetiming.data
    :members:
    :undoc-members:
    :show-inheritance:

"""



from typing import (
    Any,
    Dict,
    List,
    Sequence,
    Union
)

import matplotlib.axes

from fastf1.core import Session
from fastf1.plotting._backend import _load_drivers_from_f1_livetiming
from fastf1.plotting._base import (
    _Driver,
    _DriverTeamMapping,
    _fuzzy_matcher,
    _normalize_string,
    _Team
)
from fastf1.plotting._constants import Constants
from fastf1.plotting._constants import Constants as _Constants


_DRIVER_TEAM_MAPPINGS = dict()


def _get_driver_team_mapping(
        session: Session
) -> "_DriverTeamMapping":
    # driver-team mappings are generated once for each session and then reused
    # on future calls
    api_path = session.api_path
    year = str(session.event['EventDate'].year)

    if api_path not in _DRIVER_TEAM_MAPPINGS:
        teams = _load_drivers_from_f1_livetiming(
            api_path=api_path, year=year
        )
        mapping = _DriverTeamMapping(year, teams)
        _DRIVER_TEAM_MAPPINGS[api_path] = mapping

    return _DRIVER_TEAM_MAPPINGS[api_path]


def _get_driver(identifier: str, session: Session) -> _Driver:
    dtm = _get_driver_team_mapping(session)
    identifier = _normalize_string(identifier).lower()

    # try driver abbreviation first
    if (abb := identifier.upper()) in dtm.drivers_by_abbreviation:
        return dtm.drivers_by_abbreviation[abb]

    # check for an exact driver name match
    if identifier in dtm.drivers_by_normalized:
        return dtm.drivers_by_normalized[identifier]

    # check for exact partial string match
    for normalized_driver in dtm.drivers_by_normalized.keys():
        if identifier in normalized_driver:
            return dtm.drivers_by_normalized[normalized_driver]

    # do fuzzy string matching
    normalized_driver = _fuzzy_matcher(identifier,
                                       list(dtm.drivers_by_normalized.keys()))
    return dtm.drivers_by_normalized[normalized_driver]


def _get_team(identifier: str, session: Session) -> _Team:
    dtm = _get_driver_team_mapping(session)
    identifier = _normalize_string(identifier).lower()

    # check for an exact team name match
    if identifier in dtm.teams_by_normalized.keys():
        return dtm.teams_by_normalized[identifier]

    # check full match with full team name or for exact partial string
    # match with normalized team name
    for normalized, full in dtm.teams_by_normalized.items():
        if (identifier == full) or (identifier in normalized):
            return dtm.teams_by_normalized[normalized]

    # do fuzzy string match
    team_name = _fuzzy_matcher(identifier,
                               list(dtm.teams_by_normalized.keys()))
    return dtm.teams_by_normalized[team_name]


def _get_driver_color(
        identifier: str,
        session: Session,
        *,
        colormap: str = 'default',
        _variants: bool = False
) -> str:
    driver = _get_driver(identifier, session)
    team_name = driver.team.normalized_value

    return _get_team_color(team_name, session, colormap=colormap)


def _get_team_color(
            identifier: str,
            session: Session,
            *,
            colormap: str = 'default',
) -> str:
    dtm = _get_driver_team_mapping(session)

    if dtm.year not in Constants.keys():
        raise ValueError(f"No team colors for year '{dtm.year}'")

    team_name = _get_team(identifier, session).normalized_value
    team_consts = Constants[dtm.year].Teams[team_name]

    if colormap == 'default':
        return team_consts.TeamColor.Default
    elif colormap == 'official':
        return team_consts.TeamColor.Official
    else:
        raise ValueError(f"Invalid colormap '{colormap}'")


def get_team_name(
        identifier: str,
        session: Session,
        *,
        short: bool = False
) -> str:
    """
    Get a full team name based on a recognizable and identifiable part of
    the team name.

    Alternatively, a shortened version of the team name can be returned. The
    short version is intended for saving space when annotating plots and may
    skip parts of the official team name, for example "Scuderia Ferrari"
    becomes just "Ferrari".

    Args:
        identifier: a recognizable part of the team name
        session: the session for which the data should be obtained
        short: if True, a shortened version of the team name will be returned
    """
    team = _get_team(identifier, session)

    if short:
        dtm = _get_driver_team_mapping(session)
        team_consts = Constants[dtm.year].Teams[team.normalized_value]
        return team_consts.ShortName

    return team.value


def get_team_name_by_driver(
        identifier: str,
        session: Session,
        *,
        short: bool = False,
) -> str:
    """
    Get a full team name based on a driver's abbreviation or based on a
    recognizable and identifiable part of a driver's name.

    Alternatively, a shortened version of the team name can be returned. The
    short version is intended for saving as much space as possible when
    annotating plots and may skip parts of the official team name.

    Args:
        identifier: driver abbreviation or recognizable part of the driver name
        session: the session for which the data should be obtained
        short: if True, a shortened version of the team name will be returned
    """
    driver = _get_driver(identifier, session)
    team = driver.team

    if short:
        dtm = _get_driver_team_mapping(session)
        team_consts = Constants[dtm.year].Teams[team.normalized_value]
        return team_consts.ShortName

    return team.value


def get_team_color(
        identifier: str,
        session: Session,
        *,
        colormap: str = 'default',
) -> str:
    """
    Get a team color based on a recognizable and identifiable part of
    the team name.

    The team color is returned as a hexadecimal RGB color code.

    Args:
        identifier: a recognizable part of the team name
        session: the session for which the data should be obtained
        colormap: one of ``'default'`` or ``'official'``

    Returns:
        A hexadecimal RGB color code
    """
    return _get_team_color(identifier, session, colormap=colormap)


def get_driver_name(identifier: str, session: Session) -> str:
    """
    Get a full driver name based on the driver's abbreviation or based on
    a recognizable and identifiable part of the driver's name.

    Args:
        identifier: driver abbreviation or recognizable part of the driver name
        session: the session for which the data should be obtained
    """
    driver = _get_driver(identifier, session)
    return driver.value


def get_driver_abbreviation(identifier, session: Session) -> str:
    """
    Get a driver's abbreviation based on a recognizable and identifiable
    part of the driver's name.

    Note that the driver's abbreviation, if given exactly, is also a valid
    identifier. In this case the same value is returned as was given as the
    identifier.

    Args:
        identifier: recognizable part of the driver's name (or the
            driver's abbreviation)
        session: the session for which the data should be obtained
    """
    driver = _get_driver(identifier, session)
    return driver.abbreviation


def get_driver_names_by_team(identifier: str, session: Session) -> List[str]:
    """
    Get a list of full names of all drivers that drove for a team in a given
    session based on a recognizable and identifiable part of the team name.

    Args:
        identifier: a recognizable part of the team name
        session: the session for which the data should be obtained
    """
    team = _get_team(identifier, session)
    return [driver.value for driver in team.drivers]


def get_driver_abbreviations_by_team(
        identifier: str, session: Session
) -> List[str]:
    """
    Get a list of abbreviations of all drivers that drove for a team in a given
    session based on a recognizable and identifiable part of the team name.

    Args:
        identifier: a recognizable part of the team name
        session: the session for which the data should be obtained
    """
    team = _get_team(identifier, session)
    return [driver.abbreviation for driver in team.drivers]


def get_driver_color(
        identifier: str,
        session: Session,
        *,
        colormap: str = 'default',
) -> str:
    """
    Get the color that is associated with a driver based on the driver's
    abbreviation or based on a recognizable and identifiable part of the
    driver's name.

    .. note:: This will simply return the team color of the team that the
        driver participated for in this session. Contrary to older versions
        of FastF1, there are no separate colors for each driver. This was
        changed, because the small differences in color between drivers of
        the same team can be barely distinguished in a plot. You should use
        styling options apart from color if you need to differentiate drivers
        of the same team. The function
        :func:`~fastf1.plotting.get_driver_style` can help you to customize
        the plot styling for each driver.

    Args:
        identifier: driver abbreviation or recognizable part of the driver name
        session: the session for which the data should be obtained
        colormap: one of ``'default'`` or ``'official'``

    Returns:
        A hexadecimal RGB color code

    """
    return _get_driver_color(identifier, session, colormap=colormap)


def get_driver_style(
        identifier: str,
        style: Union[str, Sequence[str], Sequence[dict]],
        session: Session,
        *,
        colormap: str = 'default',
        additional_color_kws: Union[list, tuple] = ()
) -> Dict[str, Any]:
    """
    Get a plotting style that is unique for a driver based on the driver's
    abbreviation or based on a recognizable and identifiable part of the
    driver's name.

    This function simplifies the task of generating unique and easily
    distinguishable visual styles for multiple drivers in a plot.
    Primarily, the focus is on plotting with Matplotlib, but it is possible
    to customize the behaviour for compatibility with other plotting libraries.

    The general idea for creating visual styles is as follows:

    1. Set the primary color of the style to the color of the team for
       which a driver is driving. This may be the line color in a line plot,
       the marker color in a scatter plot, and so on.

    2. Use one or multiple other styling options (line style, markers, ...)
       to differentiate drivers in the same team.

    .. note:: It cannot be guaranteed that the styles are consistent throughout
        a full season, especially in case of driver changes within a team.


    **Option 1**: Rely on built-in styling options

    By default, this function supports the following Matplotlib plot arguments:
    ``linestyle``, ``marker``, ``color``, ``facecolor``, ``edgecolor`` as well
    as almost all other color-related arguments.

    The styling options include one color for each team and up to four
    different line styles and marker styles within a team. That means that no
    more than four different drivers are supported for a team in a single
    session. This should be sufficent in almost all scenarios.

    The following example obtains the driver style for Alonso and Stroll in a
    race in the 2023 season. The drivers should be represented using the
    ``color`` and ``marker`` arguments, as may be useful in a scatter plot.
    Both drivers were driving for the Aston Martin team, therefore, both
    automatically get assigned the same color, which is the Aston Martin team
    color. But both drivers get assigned a different marker style, so they can
    be uniquely identified in the plot.

    Example::
        >>> from fastf1 import get_session
        >>> from fastf1.plotting import get_driver_style
        >>> session = get_session(2023, 10, 'R')
        >>> get_driver_style('ALO', ['color', 'marker'], session)
        {'color': '#00665e', 'marker': 'x'}
        >>> get_driver_style('STR', ['color', 'marker'], session)
        {'color': '#00665e', 'marker': 'o'}

    **Option 2**: Provide a custom list of styling variants

    To allow for almost unlimited styling options, it is possible to specify
    custom styling variants. These are not tied to any specific plotting
    library.

    In the following example, a list with two custom stlyes is defined that are
    then used to generate driver specific styles. Each style is represented by
    a dictionary of keywords and values.

    The first driver in a team gets assigned the first style, the second driver
    the second style and so on (if there are more than two drivers). It is
    necessary to define at least as many styles as there are drivers in a team.

    The following things need to be noted:

    1. The notion of first or second driver does not refer to any particular
    reference and no specific order for drivers within a team is intended or
    guranteed.

    2. Any color-related key can make use of the "magic" ``'auto'`` value as
    shown with Alonso in this example. When the color value is set to
    ``'auto'`` it will automatically be replaced with the team color for this
    driver. All color keywords that are used in Matplotlib should be recognized
    automatically. You can define custom arguments as color arguments through
    the ``additional_color_kws`` argument.

    3. Each style dictionary can contain arbitrary keys and value. Therefore,
    you are not limited to any particular plotting library.

    Example::
        >>> from fastf1 import get_session
        >>> from fastf1.plotting import get_driver_style
        >>> session = get_session(2023, 10, 'R')
        >>> my_styles = [ \
                {'linestyle': 'solid', 'color': 'auto', 'custom_arg': True}, \
                {'linestyle': 'dotted', 'color': '#FF0060', 'other_arg': 10} \
            ]
        >>> get_driver_style('ALO', my_styles, session)
        {'linestyle': 'solid', 'color': '#00665e', 'custom_arg': True}
        >>> get_driver_style('STR', my_styles, session)
        {'linestyle': 'dotted', 'color': '#FF0060', 'other_arg': 10}

    Args:
        identifier: driver abbreviation or recognizable part of the driver name
        style: list of matplotlib plot arguments that should be used for
            styling or a list of custom style dictionaries
        session: the session for which the data should be obtained
        colormap: one of ``'default'`` or ``'official'``
        additional_color_kws: A list of keys that should additionally be
            treated as colors. This is most usefull for making the magic
            ``'auto'`` color work with custom styling options.

    Returns: a dictionary of plot style arguments that can be directly passed
        to a matplotlib plot function using the ``**`` expansion operator


    .. minigallery:: fastf1.plotting.get_driver_style
        :add-heading:
    """
    stylers = {
        'linestyle': ['solid', 'dashed', 'dashdot', 'dotted'],
        'marker': ['x', 'o', '^', 'D']
    }

    # color keyword arguments that are supported by various matplotlib
    # functions
    color_kwargs = (
        # generic
        'color', 'colors', 'c',
        # .plot
        'gapcolor',
        'markeredgecolor', 'mec',
        'markerfacecolor', 'mfc',
        'markerfacecoloralt', 'mfcalt',
        # .scatter
        'facecolor', 'facecolors', 'fc',
        'edgecolor', 'edgecolors', 'ec',
        # .errorbar
        'ecolor',
        # add user defined color keyword arguments
        *additional_color_kws
    )

    driver = _get_driver(identifier, session)
    team = driver.team
    idx = team.drivers.index(driver)

    if not style:
        # catches empty list, tuple, str
        raise ValueError("The provided style info is empty!")

    if isinstance(style, str):
        style = [style]

    plot_style = dict()

    if isinstance(style[0], str):
        # generate the plot style based on the provided keyword
        # arguments
        for opt in style:
            if opt in color_kwargs:
                value = _get_team_color(team.normalized_value,
                                        session,
                                        colormap=colormap)
            elif opt in stylers:
                value = stylers[opt][idx]
            else:
                raise ValueError(f"'{opt}' is not a supported styling "
                                 f"option")
            plot_style[opt] = value

    else:
        try:
            custom_style = style[idx]
        except IndexError:
            raise ValueError(f"The provided custom style info does not "
                             f"contain enough variants! (Has: {len(style)}, "
                             f"Required: {idx})")

        if not isinstance(custom_style, dict):
            raise ValueError("The provided style info has an invalid format!")

        # copy the correct user provided style and replace any 'auto'
        # colors with the correct color value
        plot_style = custom_style.copy()
        for kwarg in color_kwargs:
            if plot_style.get(kwarg, None) == 'auto':
                color = _get_team_color(team.normalized_value,
                                        session,
                                        colormap=colormap)
                plot_style[kwarg] = color

    return plot_style


def get_compound_color(compound: str, session: Session) -> str:
    """
    Get the compound color as hexadecimal RGB color code for a given compound.

    Args:
        compound: The name of the compound
        session: the session for which the data should be obtained

    Returns:
        A hexadecimal RGB color code
    """
    year = str(session.event['EventDate'].year)
    return _Constants[year].CompoundColors[compound.upper()]


def get_compound_mapping(session: Session) -> Dict[str, str]:
    """
    Returns a dictionary that maps compound names to their associated
    colors. The colors are given as hexadecimal RGB color codes.

    Args:
        session: the session for which the data should be obtained

    Returns:
        dictionary mapping compound names to RGB hex colors
    """
    year = str(session.event['EventDate'].year)
    return Constants[year].CompoundColors.copy()


def get_driver_color_mapping(
        session: Session, *, colormap: str = 'default',
) -> Dict[str, str]:
    """
    Returns a dictionary that maps driver abbreviations to their associated
    colors. The colors are given as hexadecimal RGB color codes.

    Args:
        session: the session for which the data should be obtained
        colormap: one of ``'default'`` or ``'official'``

    Returns:
        dictionary mapping driver abbreviations to RGB hex colors
    """
    dtm = _get_driver_team_mapping(session)

    if colormap == 'default':
        colors = {
            abb: (Constants[dtm.year]
                  .Teams[driver.team.normalized_value]
                  .TeamColor.Default)
            for abb, driver in dtm.drivers_by_abbreviation.items()
        }
    elif colormap == 'official':
        colors = {
            abb: (Constants[dtm.year]
                  .Teams[driver.team.normalized_value]
                  .TeamColor.Official)
            for abb, driver in dtm.drivers_by_abbreviation.items()
        }
    else:
        raise ValueError(f"Invalid colormap '{colormap}'")

    return colors


def list_team_names(session: Session) -> List[str]:
    """Returns a list of full team names of all teams in the ``session``."""
    dtm = _get_driver_team_mapping(session)
    return list(team.value for team in dtm.teams_by_normalized.values())


def list_short_team_names(session: Session) -> List[str]:
    """Returns a list of short team names of all teams in the ``session``."""
    dtm = _get_driver_team_mapping(session)
    return list(Constants[dtm.year].Teams[team].ShortName
                for team in dtm.teams_by_normalized.keys())


def list_driver_abbreviations(session: Session) -> List[str]:
    """Returns a list of abbreviations of all drivers in the ``session``."""
    dtm = _get_driver_team_mapping(session)
    return list(dtm.drivers_by_abbreviation.keys())


def list_driver_names(session: Session) -> List[str]:
    """Returns a list of full names of all drivers in the ``session``."""
    dtm = _get_driver_team_mapping(session)
    return list(driver.value for driver in dtm.drivers_by_normalized.values())


def list_compounds(session: Session) -> List[str]:
    """Returns a list of all compound names for this season (not session)."""
    year = str(session.event['EventDate'].year)
    return list(Constants[year].CompoundColors.keys())


def add_sorted_driver_legend(ax: matplotlib.axes.Axes, session: Session):
    """
    Adds a legend to the axis where drivers are ordered by team and within a
    team in the same order that is used for selecting plot styles.

    This function is a drop in replacement for calling Matplotlib's
    ``ax.legend()`` method. It can only be used when driver names or driver
    abbreviations are used as labels for the legend.

    There is no particular need to use this function except to make the
    legend more visually pleasing.

    Args:
        ax: An instance of a Matplotlib ``Axes`` object
        session: the session for which the data should be obtained

     Returns:
        ``matplotlib.legend.Legend``

    .. minigallery:: fastf1.plotting.add_sorted_driver_legend
        :add-heading:

    """
    dtm = _get_driver_team_mapping(session)
    handles, labels = ax.get_legend_handles_labels()

    teams_list = list(dtm.teams_by_normalized.values())
    driver_list = list(dtm.drivers_by_normalized.values())

    # create an intermediary list where each element is a tuple that
    # contains (team_idx, driver_idx, handle, label). Then sort this list
    # based on the team_idx and driver_idx. As a result, drivers from the
    # same team will be next to each other and in the same order as their
    # styles are cycled.
    ref = list()
    for hdl, lbl in zip(handles, labels):
        driver = _get_driver(lbl, session)
        team = driver.team

        team_idx = teams_list.index(team)
        driver_idx = driver_list.index(driver)

        ref.append((team_idx, driver_idx, hdl, lbl))

    # sort based only on team_idx and driver_idx (i.e. first two entries)
    ref.sort(key=lambda e: e[:2])

    handles_new = list()
    labels_new = list()
    for elem in ref:
        handles_new.append(elem[2])
        labels_new.append(elem[3])

    return ax.legend(handles_new, labels_new)



"""
Position changes during a race
==============================

Plot the position of each driver at the end of each lap.
"""

import matplotlib.pyplot as plt

import fastf1.plotting


# Load FastF1's dark color scheme
fastf1.plotting.setup_mpl(mpl_timedelta_support=False, misc_mpl_mods=False,
                          color_scheme='fastf1')


##############################################################################
# Load the session and create the plot
session = fastf1.get_session(2023, 1, 'R')
session.load(telemetry=False, weather=False)

fig, ax = plt.subplots(figsize=(8.0, 4.9))
# sphinx_gallery_defer_figures

##############################################################################
# For each driver, get their three letter abbreviation (e.g. 'HAM') by simply
# using the value of the first lap, get their color and then plot their
# position over the number of laps.
for drv in session.drivers:
    drv_laps = session.laps.pick_driver(drv)

    abb = drv_laps['Driver'].iloc[0]
    style = fastf1.plotting.get_driver_style(identifier=abb,
                                             style=['color', 'linestyle'],
                                             session=session)

    ax.plot(drv_laps['LapNumber'], drv_laps['Position'],
            label=abb, **style)
# sphinx_gallery_defer_figures

##############################################################################
# Finalize the plot by setting y-limits that invert the y-axis so that position
# one is at the top, set custom tick positions and axis labels.
ax.set_ylim([20.5, 0.5])
ax.set_yticks([1, 5, 10, 15, 20])
ax.set_xlabel('Lap')
ax.set_ylabel('Position')
# sphinx_gallery_defer_figures

##############################################################################
# Because this plot is very crowed, add the legend outside the plot area.
ax.legend(bbox_to_anchor=(1.0, 1.02))
plt.tight_layout()

plt.show()



"""Plot driver standings in a heatmap
======================================

Plot the points for each driven in each race of a given season in a heatmap, as
https://public.tableau.com/app/profile/mateusz.karmalski/viz/F1ResultsTracker2022
"""

import pandas as pd
import plotly.express as px
from plotly.io import show

from fastf1.ergast import Ergast


##############################################################################
# First, we load the results for season 2022.
ergast = Ergast()
races = ergast.get_race_schedule(2022)  # Races in year 2022
results = []

# For each race in the season
for rnd, race in races['raceName'].items():

    # Get results. Note that we use the round no. + 1, because the round no.
    # starts from one (1) instead of zero (0)
    temp = ergast.get_race_results(season=2022, round=rnd + 1)
    temp = temp.content[0]

    # If there is a sprint, get the results as well
    sprint = ergast.get_sprint_results(season=2022, round=rnd + 1)
    if sprint.content and sprint.description['round'][0] == rnd + 1:
        temp = pd.merge(temp, sprint.content[0], on='driverCode', how='left')
        # Add sprint points and race points to get the total
        temp['points'] = temp['points_x'] + temp['points_y']
        temp.drop(columns=['points_x', 'points_y'], inplace=True)

    # Add round no. and grand prix name
    temp['round'] = rnd + 1
    temp['race'] = race.removesuffix(' Grand Prix')
    temp = temp[['round', 'race', 'driverCode', 'points']]  # Keep useful cols.
    results.append(temp)

# Append all races into a single dataframe
results = pd.concat(results)
races = results['race'].drop_duplicates()

##############################################################################
# Then we “reshape” the results to a wide table, where each row represents a
# driver and each column refers to a race, and the cell value is the points.
results = results.pivot(index='driverCode', columns='round', values='points')
# Here we have a 22-by-22 matrix (22 races and 22 drivers, incl. DEV and HUL)

# Rank the drivers by their total points
results['total_points'] = results.sum(axis=1)
results = results.sort_values(by='total_points', ascending=False)
results.drop(columns='total_points', inplace=True)

# Use race name, instead of round no., as column names
results.columns = races


##############################################################################
# The final step is to plot a heatmap using plotly
fig = px.imshow(
    results,
    text_auto=True,
    aspect='auto',  # Automatically adjust the aspect ratio
    color_continuous_scale=[[0,    'rgb(198, 219, 239)'],  # Blue scale
                            [0.25, 'rgb(107, 174, 214)'],
                            [0.5,  'rgb(33,  113, 181)'],
                            [0.75, 'rgb(8,   81,  156)'],
                            [1,    'rgb(8,   48,  107)']],
    labels={'x': 'Race',
            'y': 'Driver',
            'color': 'Points'}       # Change hover texts
)
fig.update_xaxes(title_text='')      # Remove axis titles
fig.update_yaxes(title_text='')
fig.update_yaxes(tickmode='linear')  # Show all ticks, i.e. driver names
fig.update_yaxes(showgrid=True, gridwidth=1, gridcolor='LightGrey',
                 showline=False,
                 tickson='boundaries')              # Show horizontal grid only
fig.update_xaxes(showgrid=False, showline=False)    # And remove vertical grid
fig.update_layout(plot_bgcolor='rgba(0,0,0,0)')     # White background
fig.update_layout(coloraxis_showscale=False)        # Remove legend
fig.update_layout(xaxis=dict(side='top'))           # x-axis on top
fig.update_layout(margin=dict(l=0, r=0, b=0, t=0))  # Remove border margins
fig
show(fig)



"""
Event Schedule - :mod:`fastf1.events`
=====================================

The :class:`EventSchedule` provides information about past and upcoming
Formula 1 events.

An :class:`Event` can be a race weekend or a testing event. Each event
consists of multiple :class:`~fastf1.core.Session`.

The event schedule objects are built on top of pandas'
:class:`pandas.DataFrame` (event schedule) and :class:`pandas.Series` (event).
Therefore, the usual methods of these pandas objects can be used in addition
to the special methods described here.

Event Schedule Data
-------------------

The event schedule and each event provide the following information as
DataFrame columns or Series values:

  - ``RoundNumber`` | :class:`int` |
    The number of the championship round. This is unique for race
    weekends, while testing events all share the round number zero.

  - ``Country`` | :class:`str` | The country in which the event is held.

  - ``Location`` | :class:`str` |
    The event location; usually the city or region in which the track is
    situated.

  - ``OfficialEventName`` | :class:`str` |
    The official event name as advertised, including sponsor names and stuff.

  - ``EventName`` | :class:`str` |
    A shorter event name usually containing the country or location but no
    no sponsor names. This name is required internally for proper api access.

  - ``EventDate`` | :class:`pd.Timestamp` |
    The events reference date and time. This is used mainly internally.
    Usually, this is the same as the date of the last session.

  - ``EventFormat`` | :class:`str` |
    The format of the event. One of 'conventional', 'sprint',
    'sprint_shootout', 'sprint_qualifying', 'testing'. See :ref:`event-formats`

  - ``Session*`` | :class:`str` |
    The name of the session. One of 'Practice 1', 'Practice 2', 'Practice 3',
    'Qualifying', 'Sprint', 'Sprint Shootout' or 'Race'.
    Testing sessions are considered practice.
    ``*`` denotes the number of
    the session (1, 2, 3, 4, 5).

  - ``Session*Date`` | :class:`pd.Timestamp` |
    The date and time at which the session is scheduled to start or was
    scheduled to start as timezone-aware local timestamp.
    (Timezone-aware local timestamps are not available when the ``'ergast'``
    backend is used.)
    ``*`` denotes the number of the session (1, 2, 3, 4, 5).

  - ``Session*DateUtc`` | :class:`pd.Timestamp` |
    The date and time at which the session is scheduled to start or was
    scheduled to start as non-timezone-aware UTC timestamp.
    ``*`` denotes the number of the session (1, 2, 3, 4, 5).

  - ``F1ApiSupport`` | :class:`bool` |
    Denotes whether this session is supported by the official F1 API.
    Lap timing data and telemetry data can only be loaded if this is true.


Supported Seasons
.................

FastF1 provides its own event schedule for the 2018 season and all later
seasons. The schedule for all seasons before 2018 is built using data from
the Ergast API. Only limited data is available for these seasons. Usage of the
Ergast API can be enforced for all seasons by setting ``backend='ergast'``,
in which case the same limitations apply for the more recent seasons too.

**Exact scheduled starting times for all sessions**:
Supported starting with the 2018 season.
Starting dates for sessions before 2018 (or when enforcing usage of the Ergast
API) assume that each race weekend was held according to the 'conventional'
schedule (Practice 1/2 on friday, Practice 3/Qualifying on Saturday, Race on
Sunday). A starting date and time can only be provided for the race session.
All other sessions are calculated from this and no starting times can be
provided for these. These assumptions will be incorrect for certain events!

**Testing events**: Supported for the 2020 season and later seasons. Not
supported if usage of the Ergast API is enforced.


.. _event-formats:

Event Formats
.............

- **'conventional': Practice 1, Practice 2, Practice 3, Qualifying, Race**

- **'sprint': Practice 1, Qualifying, Practice 2, Sprint, Race**

  This is the original sprint format that was used in some races in 2021 and
  2022. The Qualifying on friday set the grid for the Sprint on saturday.
  The results from the Sprint set the grid for the Race on sunday.

- **'sprint_shootout': Practice 1, Qualifying, Sprint Shootout, Sprint, Race**

  This format was used in 2023. The Qualifying on friday sets the grid for the
  main Race on sunday. The Sprint Shootout on saturday is held in similar
  fashion to a normal Qualifying session and sets the grid for the Sprint that
  takes place on saturday as well.

- **'sprint_qualifying': Practice 1, Sprint Qualifying, Sprint, Qualifying,
  Race**

  This format is used starting from 2024. In general, it is similar to the
  previous 'sprint_shootout' format, but the order of the sessions was changed
  and 'Sprint Shootout' is renamed to 'Sprint Qualifying'. This means that
  the Sprint Qualifying on friday is held in similar fashion to a normal
  Qualifying and sets the grid for the Sprint on saturday. The Qualifying later
  on saturday then sets the grid for the race on Sunday.

- **'testing': no fixed session order**

  usually three practice sessions on three separate days


.. _SessionIdentifier:

Session identifiers
-------------------

Multiple event (schedule) related functions and methods make use of a session
identifier to differentiate between the various sessions of one event.
This identifier can currently be one of the following:

    - session name abbreviation: ``'FP1', 'FP2', 'FP3', 'Q', 'S', 'SS', 'SQ',
      'R'``
    - full session name: ``'Practice 1', 'Practice 2',
      'Practice 3', 'Sprint', 'Sprint Shootout', 'Sprint Qualifying',
      'Qualifying', 'Race'``;
      provided names will be normalized, so that the name is
      case-insensitive
    - number of the session: ``1, 2, 3, 4, 5``

Note that the old ``'sprint'`` event format from 2021 and 2022 originally used
the name 'Sprint Qualifying' before renaming these sessions to just 'Sprint'.
The official schedule for 2021 now lists all these sessions as 'Sprint' and
FastF1 will therefore return all these session as 'Sprint'. When querying for a
specific session, FastF1 will also accept the 'Sprint Qualifying'/'SQ'
identifier instead of only 'Sprint'/'S' for backwards compatibility.


Functions for accessing schedule data
-------------------------------------

The functions for accessing event schedule data are documented in
:ref:`GeneralFunctions`.


Data Objects
------------


Overview
........


.. autosummary::
    EventSchedule
    Event


API Reference
.............


.. autoclass:: EventSchedule
    :members:
    :undoc-members:
    :autosummary:


.. autoclass:: Event
    :members:
    :undoc-members:
    :autosummary:

"""  # noqa: W605 invalid escape sequence (escaped space)
import collections
import datetime
import json
import warnings
from typing import (
    Literal,
    Optional,
    Type,
    Union
)

import dateutil.parser


with warnings.catch_warnings():
    warnings.filterwarnings(
        'ignore', message="Using slow pure-python SequenceMatcher"
    )
    # suppress that warning, it's confusing at best here, we don't need fast
    # sequence matching and the installation (on windows) requires some effort
    from rapidfuzz import fuzz

import pandas as pd

import fastf1._api
import fastf1.ergast
from fastf1.core import Session
from fastf1.internals.pandas_base import (
    BaseDataFrame,
    BaseSeries
)
from fastf1.logger import (
    get_logger,
    soft_exceptions
)
from fastf1.req import Cache
from fastf1.utils import (
    recursive_dict_get,
    to_datetime,
    to_timedelta
)


_logger = get_logger(__name__)

_SESSION_TYPE_ABBREVIATIONS = {
    'R': 'Race',
    'Q': 'Qualifying',
    'S': 'Sprint',
    'SQ': 'Sprint Qualifying',
    'SS': 'Sprint Shootout',
    'FP1': 'Practice 1',
    'FP2': 'Practice 2',
    'FP3': 'Practice 3'
}

_SCHEDULE_BASE_URL = "https://raw.githubusercontent.com/" \
                     "theOehrly/f1schedule/master/"


def get_session(
        year: int,
        gp: Union[str, int],
        identifier: Optional[Union[int, str]] = None,
        *,
        backend: Optional[Literal['fastf1', 'f1timing', 'ergast']] = None,
        force_ergast: bool = False,
) -> Session:
    """Create a :class:`~fastf1.core.Session` object based on year, event name
    and session identifier.

    .. note:: This function will return a :class:`~fastf1.core.Session`
        object, but it will not load any session specific data like lap timing,
        telemetry, ... yet. For this, you will need to call
        :func:`~fastf1.core.Session.load` on the returned object.

    To get a testing session, use :func:`get_testing_session`.

    Examples:

        Get the second free practice of the first race of 2021 by its session
        name abbreviation::

            >>> get_session(2021, 1, 'FP2')

        Get the qualifying of the 2020 Austrian Grand Prix by full session
        name::

            >>> get_session(2020, 'Austria', 'Qualifying')

        Get the 3rd session of the 5th Grand Prix in 2021::

            >>> get_session(2021, 5, 3)

    Args:
        year: Championship year

        gp: Name as str or round number as int. If gp is
            a string, a fuzzy match will be performed on all events and the
            closest match will be selected.
            Fuzzy matching uses country, location, name and officialName of
            each event as reference.

            Some examples that will be correctly interpreted: 'bahrain',
            'australia', 'abudabi', 'monza'.

        identifier: see :ref:`SessionIdentifier`

        backend: select a specific backend as data source, options:
            - ``'fastf1'``: FastF1's own backend, full support for 2018 to now

            - ``'f1timing'``: uses data from the F1 live timing API, sessions
              for which no timing data is available are not listed
              (supports 2018 to now)

            - ``'ergast'``: uses data from Ergast, no local times are
              available, no information about availability of f1 timing data is
              available (supports 1950 to now)

            When no backend is specified, ``'fastf1'`` is used as a default and
            the other backends are used as a fallback in case the default
            is not available.

            For seasons older than 2018 ``'ergast'`` is always used.

        force_ergast: [Deprecated, use ``backend='ergast'``] Always use data
            from the ergast database to create the event schedule
    """
    event = get_event(year, gp, force_ergast=force_ergast, backend=backend)
    return event.get_session(identifier)


def get_testing_session(
        year: int,
        test_number: int,
        session_number: int,
        *,
        backend: Optional[Literal['fastf1', 'f1timing']] = None
) -> Session:
    """Create a :class:`~fastf1.core.Session` object for testing sessions
    based on year, test  event number and session number.

    Args:
        year: Championship year

        test_number: Number of the testing event (usually at most two)

        session_number: Number of the session within a specific testing
            event. Each testing event usually has three sessions.

        backend: select a specific backend as data source, options:

            - ``'fastf1'``: FastF1's own backend, full support for 2018 to now

            - ``'f1timing'``: uses data from the F1 live timing API, sessions
                for which no timing data is available are not listed
                (supports 2018 to now)

            When no backend is specified, ``'fastf1'`` is used as a default and
            ``f1timing`` is used as a fallback in case the default
            is not available.

    .. versionadded:: 2.2
    """
    event = get_testing_event(year, test_number, backend=backend)
    return event.get_session(session_number)


def get_event(
        year: int,
        gp: Union[int, str],
        *,
        backend: Optional[Literal['fastf1', 'f1timing', 'ergast']] = None,
        force_ergast: bool = False,
        strict_search: bool = False,
        exact_match: bool = False
) -> "Event":
    """Create an :class:`~fastf1.events.Event` object for a specific
    season and gp.

    To get a testing event, use :func:`get_testing_event`.

    Args:
        year: Championship year

        gp: Name as str or round number as int. If gp is
            a string, a fuzzy match will be performed on all events and the
            closest match will be selected.
            Fuzzy matching uses country, location, name and officialName of
            each event as reference.
            Note that the round number cannot be used to get a testing event,
            as all testing event are round 0!

        backend: select a specific backend as data source, options:

            - ``'fastf1'``: FastF1's own backend, full support for 2018 to now

            - ``'f1timing'``: uses data from the F1 live timing API, sessions
              for which no timing data is available are not listed
              (supports 2018 to now)

            - ``'ergast'``: uses data from Ergast, no local times are
              available, no information about availability of f1 timing data is
              available (supports 1950 to now)

            When no backend is specified, ``'fastf1'`` is used as a default and
            the other backends are used as a fallback in case the default
            is not available.

            For seasons older than 2018 ``'ergast'`` is always used.

        force_ergast: [Deprecated, use ``backend='ergast'``] Always use data
            from the ergast database to create the event schedule

        strict_search: This argument is deprecated and planned for removal,
            use the equivalent ``exact_match`` instead

        exact_match: Match precisely the query, or default to
            fuzzy search. If no event is found with
            ``exact_match=True``, the function will return None

    .. versionadded:: 2.2
    """
    schedule = get_event_schedule(year=year, include_testing=False,
                                  force_ergast=force_ergast,
                                  backend=backend)

    if isinstance(gp, str):
        event = schedule.get_event_by_name(
            gp, strict_search=strict_search, exact_match=exact_match)
    else:
        event = schedule.get_event_by_round(gp)

    return event


def get_testing_event(
        year: int,
        test_number: int,
        *,
        backend: Optional[Literal['fastf1', 'f1timing']] = None
) -> "Event":
    """Create a :class:`fastf1.events.Event` object for testing sessions
    based on year and test event number.

    Args:
        year: Championship year
        test_number: Number of the testing event (usually at most two)
        backend: select a specific backend as data source, options:

            - ``'fastf1'``: FastF1's own backend, full support for 2018 to now

            - ``'f1timing'``: uses data from the F1 live timing API, sessions
              for which no timing data is available are not listed
              (supports 2018 to now)

            When no backend is specified, ``'fastf1'`` is used as a default and
            ``f1timing`` is used as a fallback in case the default
            is not available.

    .. versionadded:: 2.2
    """
    if backend == 'ergast':
        raise ValueError("The 'ergast' backend does not support "
                         "testing events!")
    schedule = get_event_schedule(year=year, backend=backend)
    schedule = schedule[schedule.is_testing()]

    try:
        assert test_number >= 1
        return schedule.iloc[test_number - 1]
    except (IndexError, AssertionError):
        raise ValueError(f"Test event number {test_number} does not exist")


def get_event_schedule(
        year: int,
        *,
        include_testing: bool = True,
        backend: Optional[Literal['fastf1', 'f1timing', 'ergast']] = None,
        force_ergast: bool = False
) -> "EventSchedule":
    """Create an :class:`~fastf1.events.EventSchedule` object for a specific
    season.

    Args:
        year: Championship year

        include_testing: Include or exclude testing sessions from the
            event schedule.

        backend: select a specific backend as data source, options:

            - ``'fastf1'``: FastF1's own backend, full support for 2018 to now

            - ``'f1timing'``: uses data from the F1 live timing API, sessions
              for which no timing data is available are not listed
              (supports 2018 to now)

            - ``'ergast'``: uses data from Ergast, no local times are
              available, no information about availability of f1 timing data is
              available (supports 1950 to now)

            When no backend is specified, ``'fastf1'`` is used as a default and
            the other backends are used as a fallback in case the default
            is not available.

            For seasons older than 2018 ``'ergast'`` is always used.

        force_ergast: [Deprecated, use ``backend='ergast'``] Always use data
            from the ergast database to create the event schedule

    .. versionadded:: 2.2

    """
    if force_ergast:
        warnings.warn("Option ``force_ergast`` has been deprecated, use"
                      "``backend='ergast'`` instead")
        backend = 'ergast'

    _backends_named_order = {
        'fastf1': _get_schedule_ff1,
        'f1timing': _get_schedule_from_f1_timing,
        'ergast': _get_schedule_from_ergast
    }

    if backend is not None:
        _backends = [_backends_named_order[backend]]
    elif year < 2018:
        _backends = [_backends_named_order['ergast']]
    else:
        _backends = list(_backends_named_order.values())

    schedule = None
    for func in _backends:
        schedule = func(year)
        if schedule is not None:
            break

    if schedule is None:  # raise Error if fallback failed as well
        raise ValueError("Failed to load any schedule data.")

    if not include_testing:
        schedule = schedule[~schedule.is_testing()]
    return schedule


def get_events_remaining(
        dt: Optional[datetime.datetime] = None,
        *,
        include_testing: bool = True,
        backend: Optional[Literal['fastf1', 'f1timing', 'ergast']] = None,
        force_ergast: bool = False
) -> 'EventSchedule':
    """
    Create an :class:`~fastf1.events.EventSchedule` object for remaining
    season.

    Args:
        dt: Optional DateTime to get events after.

        include_testing: Include or exclude testing sessions from the
            event schedule.

        backend: select a specific backend as data source, options:

            - ``'fastf1'``: FastF1's own backend, full support for 2018 to now

            - ``'f1timing'``: uses data from the F1 live timing API, sessions
              for which no timing data is available are not listed
              (supports 2018 to now)

            - ``'ergast'``: uses data from Ergast, no local times are
              available, no information about availability of f1 timing data is
              available (supports 1950 to now)

            When no backend is specified, ``'fastf1'`` is used as a default and
            the other backends are used as a fallback in case the default
            is not available.

            For seasons older than 2018 ``'ergast'`` is always used.

        force_ergast: [Deprecated, use ``backend='ergast'``] Always use data
            from the ergast database to create the event schedule

    .. versionadded:: 2.3
    """
    if dt is None:
        dt = datetime.datetime.now()

    events = get_event_schedule(
        dt.year, include_testing=include_testing,
        force_ergast=force_ergast, backend=backend
    )
    result = events.loc[events["EventDate"] >= dt]
    return result


@soft_exceptions("FastF1 schedule",
                 "Failed to load schedule from FastF1 backend!",
                 _logger)
def _get_schedule_ff1(year):
    response = Cache.requests_get(
        _SCHEDULE_BASE_URL + f"schedule_{year}.json"
    )

    data = dict()
    json_data = json.loads(response.text)
    for key in json_data.keys():
        data[key] = list(json_data[key].values())

    # convert gmt offset to timedelta
    gmt_offset = list()
    for go in data.pop('gmt_offset'):
        if go is None:
            gmt_offset.append(datetime.timedelta(0))
        else:
            # hh:mm -> hh:mm:00 before to_timedelta
            gmt_offset.append(to_timedelta(f"{go}:00"))

    length = len(gmt_offset)

    # create additional columns for UTC timestamps
    for n in range(5):
        data[f'session{n+1}_date_Utc'] = [None, ] * length

    # convert and set all timestamps
    for i in range(length):
        data['event_date'][i] = to_datetime(data['event_date'][i]) \
            .replace(hour=0, minute=0, second=0)

        for j in range(5):
            date = to_datetime(data[f'session{j+1}_date'][i])
            if date is None:
                date_utc = None
            else:
                # create tz-aware local time
                date = date.replace(tzinfo=datetime.timezone(gmt_offset[i]))

                # create non-tz-aware utc time
                date_utc = date.astimezone(datetime.timezone.utc) \
                    .replace(tzinfo=None)

            data[f'session{j+1}_date'][i] = pd.Timestamp(date)
            data[f'session{j+1}_date_Utc'][i] = pd.Timestamp(date_utc)

    str().capitalize()

    df = pd.DataFrame(data)
    # change column names from snake_case to UpperCamelCase
    col_renames = {col: ''.join([s.capitalize() for s in col.split('_')])
                   for col in df.columns}
    df = df.rename(columns=col_renames)

    schedule = EventSchedule(df, year=year, force_default_cols=True)
    return schedule


@soft_exceptions("F1 API schedule",
                 "Failed to load schedule from F1 API backend!",
                 _logger)
def _get_schedule_from_f1_timing(year: int):
    # create an event schedule using data from the F1 API
    response = fastf1._api.season_schedule(f'/static/{year}/')
    data = collections.defaultdict(list)

    for event in response:
        data['Country'].append(event['Country']['Name'])
        data['Location'].append(event['Location'])
        data['EventName'].append(event['Name'])
        data['OfficialEventName'].append(event['OfficialName'])

        # select only valid sessions
        sessions = list()
        for ses in event['Sessions']:
            if (ses.get('Key') != -1) and ses.get('Name'):
                sessions.append(ses)

        n_events = min(len(sessions), 5)
        # number of events, usually 3 for testing, 5 for race weekends
        # in special cases there are additional unrelated events

        if 'test' in event['Name'].lower():
            data['EventFormat'].append('testing')
            data['RoundNumber'].append(0)

        elif year <= 2020:
            data['EventFormat'].append('conventional')
            data['RoundNumber'].append(event['Number'])

        elif year in (2021, 2022):
            if sessions[3]['Name'] == 'Sprint Qualifying':
                # fix for 2021 where Sprint was called Sprint Qualifying
                sessions[3]['Name'] = 'Sprint'

            if sessions[3]['Name'] == 'Sprint':
                data['EventFormat'].append('sprint')
            else:
                data['EventFormat'].append('conventional')
            data['RoundNumber'].append(event['Number'])

        elif year == 2023:
            if sessions[2]['Name'] == 'Sprint Shootout':
                data['EventFormat'].append('sprint_shootout')
            else:
                data['EventFormat'].append('conventional')
            data['RoundNumber'].append(event['Number'])

        elif year >= 2024:
            if sessions[1]['Name'] == 'Sprint Qualifying':
                data['EventFormat'].append('sprint_qualifying')
            else:
                data['EventFormat'].append('conventional')
            data['RoundNumber'].append(event['Number'])

        data['F1ApiSupport'].append(True)

        for i in range(0, 5):
            # parse the up to five sessions for each event
            try:
                session = sessions[i]
            except IndexError:
                data[f'Session{i+1}'].append(None)
                data[f'Session{i+1}Date'].append(None)
                data[f'Session{i+1}DateUtc'].append(None)
            else:
                data[f'Session{i+1}'].append(session['Name'])
                # save timestamp as tz-aware local time and non-tz-aware utc
                date = to_datetime(session['StartDate'])
                gmt_offset = to_timedelta(session['GmtOffset'])
                date = date.replace(tzinfo=datetime.timezone(gmt_offset))
                date_utc = date.astimezone(datetime.timezone.utc) \
                    .replace(tzinfo=None)
                data[f'Session{i+1}Date'].append(pd.Timestamp(date))
                data[f'Session{i+1}DateUtc'].append(pd.Timestamp(date_utc))

        # set the event date to the date of the last session
        ev_date = data[f'Session{n_events}DateUtc'][-1]
        ev_date = ev_date.replace(hour=0, minute=0, second=0)
        data['EventDate'].append(ev_date)

    schedule = EventSchedule(data, year=year, force_default_cols=True)
    return schedule


@soft_exceptions("Ergast API Schedule",
                 "Failed to load schedule from Ergast API backend!",
                 _logger)
def _get_schedule_from_ergast(year) -> "EventSchedule":
    # create an event schedule using data from the ergast database
    season = fastf1.ergast.fetch_season(year)
    data = collections.defaultdict(list)
    for rnd in season:
        data['RoundNumber'].append(int(rnd.get('round')))
        data['Country'].append(
            recursive_dict_get(rnd, 'Circuit', 'Location', 'country')
        )
        data['Location'].append(
            recursive_dict_get(rnd, 'Circuit', 'Location', 'locality')
        )
        data['EventName'].append(rnd.get('raceName'))
        data['OfficialEventName'].append("")

        try:
            date = pd.to_datetime(
                f"{rnd.get('date', '')}T{rnd.get('time', '')}",
            ).tz_localize(None)
        except dateutil.parser.ParserError:
            date = pd.NaT
        data['EventDate'].append(date)

        if 'Sprint' in rnd:
            # Ergast doesn't support sprint shootout format yet
            if year in (2021, 2022):
                _format = 'sprint'
                session_names = ['Practice 1', 'Qualifying', 'Practice 2',
                                 'Sprint', 'Race']
            elif year == 2023:
                _format = 'sprint_shootout'
                session_names = ['Practice 1', 'Qualifying', 'Sprint Shootout',
                                 'Sprint', 'Race']
            else:
                _format = 'sprint_qualifying'
                session_names = ['Practice 1', 'Sprint Qualifying', 'Sprint',
                                 'Qualifying', 'Race']

            data['EventFormat'].append(_format)

            data['Session1'].append(session_names[0])
            data['Session1DateUtc'].append(
                date.floor('D') - pd.Timedelta(days=2))

            data['Session2'].append(session_names[1])
            data['Session2DateUtc'].append(
                date.floor('D') - pd.Timedelta(days=2))

            data['Session3'].append(session_names[2])
            data['Session3DateUtc'].append(
                date.floor('D') - pd.Timedelta(days=1))

            data['Session4'].append(session_names[3])
            data['Session4DateUtc'].append(
                date.floor('D') - pd.Timedelta(days=1))

            data['Session5'].append(session_names[4])
            data['Session5DateUtc'].append(date)

        else:
            data['EventFormat'].append("conventional")

            data['Session1'].append('Practice 1')
            data['Session1DateUtc'].append(
                date.floor('D') - pd.Timedelta(days=2))

            data['Session2'].append('Practice 2')
            data['Session2DateUtc'].append(
                date.floor('D') - pd.Timedelta(days=2))

            data['Session3'].append('Practice 3')
            data['Session3DateUtc'].append(
                date.floor('D') - pd.Timedelta(days=1))

            data['Session4'].append('Qualifying')
            data['Session4DateUtc'].append(
                date.floor('D') - pd.Timedelta(days=1))

            data['Session5'].append('Race')
            data['Session5DateUtc'].append(date)

        data['F1ApiSupport'].append(True if year >= 2018 else False)
        # simplified; this is only true most of the time

    df = pd.DataFrame(data)
    schedule = EventSchedule(df, year=year, force_default_cols=True)
    return schedule


class EventSchedule(BaseDataFrame):
    """This class implements a per-season event schedule.

    For detailed information about the information that is available for each
    event, see `Event Schedule Data`_.

    This class is usually not instantiated directly. You should use
    :func:`fastf1.get_event_schedule` to get an event schedule for a specific
    season.

    Args:
        *args: passed on to :class:`pandas.DataFrame` superclass
        year: Championship year
        force_default_cols: Enforce that all default columns and only
            the default columns exist
        **kwargs: passed on to :class:`pandas.DataFrame` superclass
            (except 'columns' which is unsupported for the event schedule)

    .. versionadded:: 2.2
    """

    _COL_TYPES = {
        'RoundNumber': int,
        'Country': str,
        'Location': str,
        'OfficialEventName': str,
        'EventDate': 'datetime64[ns]',
        'EventName': str,
        'EventFormat': str,
        'Session1': str,
        'Session1Date': object,  # tz-aware datetime.datetime
        'Session1DateUtc': 'datetime64[ns]',
        'Session2': str,
        'Session2Date': object,
        'Session2DateUtc': 'datetime64[ns]',
        'Session3': str,
        'Session3Date': object,
        'Session3DateUtc': 'datetime64[ns]',
        'Session4': str,
        'Session4Date': object,
        'Session4DateUtc': 'datetime64[ns]',
        'Session5': str,
        'Session5Date': object,
        'Session5DateUtc': 'datetime64[ns]',
        'F1ApiSupport': bool
    }

    _metadata = ['year']

    def __init__(self, *args, year: int = 0,
                 force_default_cols: bool = False, **kwargs):
        if force_default_cols:
            kwargs['columns'] = list(self._COL_TYPES)
        super().__init__(*args, **kwargs)
        self.year = year

        # apply column specific dtypes
        for col, _type in self._COL_TYPES.items():
            if col not in self.columns:
                continue
            if self[col].isna().all():
                if _type == 'datetime64[ns]':
                    self[col] = pd.NaT
                elif _type == object:
                    self[col] = None
                else:
                    self[col] = _type()
            self[col] = self[col].astype(_type)

    @property
    def _constructor_sliced_horizontal(self) -> Type["Event"]:
        return Event

    def is_testing(self):
        """Return `True` or `False`, depending on whether each event is a
        testing event."""
        return pd.Series(self['EventFormat'] == 'testing')

    def get_event_by_round(self, round: int) -> "Event":
        """Get an :class:`Event` by its round number.

        Args:
            round: The round number
        Raises:
            ValueError: The round does not exist in the event schedule
        """
        if round == 0:
            raise ValueError("Cannot get testing event by round number!")
        mask = self['RoundNumber'] == round
        if not mask.any():
            raise ValueError(f"Invalid round: {round}")
        return self[mask].iloc[0]

    def _strict_event_search(self, name: str):
        """
        Match Event Name exactly, ignoring case.
        """

        query = name.lower()
        for i, event in self.iterrows():
            if 'EventName' in event:
                if event['EventName'].lower() == query:
                    return self.loc[i]
        else:
            return None

    def _fuzzy_event_search(self, name: str) -> "Event":

        def _remove_common_words(event_name):
            common_words = ["formula 1", str(self.year), "grand prix", "gp"]
            event_name = event_name.casefold()

            for word in common_words:
                event_name = event_name.replace(word, "")

            return event_name.replace(" ", "")

        def _matcher_strings(ev):
            strings = list()
            if 'Location' in ev:
                strings.append(ev['Location'].casefold())
            if 'Country' in ev:
                strings.append(ev['Country'].casefold())
            if 'EventName' in ev:
                strings.append(_remove_common_words(ev["EventName"]))
            if 'OfficialEventName' in ev:
                strings.append(_remove_common_words(ev["OfficialEventName"]))
            return strings

        user_input = name
        name = _remove_common_words(name)
        full_partial_match_indices = []

        # check partial matches first
        # if there is either zero or multiple 100% matches
        # fall back to the full ratio
        for i, event in self.iterrows():
            if any([name in val for val in _matcher_strings(event)]):
                full_partial_match_indices.append(i)

        if len(full_partial_match_indices) == 1:
            return self.loc[full_partial_match_indices[0]]

        max_ratio = 0
        max_index = 0

        for i, event in self.loc[full_partial_match_indices
                                  or self.index].iterrows():
            ratio = max(
                [fuzz.ratio(val, name)
                 for val in _matcher_strings(event)]
            )
            if ratio > max_ratio:
                max_ratio = ratio
                max_index = i

        if max_ratio != 100:
            _logger.warning((
                "Correcting user input "
                f"'{user_input}' to'{self.loc[max_index].EventName}'"
            )
            )
        return self.loc[max_index]

    def get_event_by_name(
            self,
            name: str,
            *,
            strict_search: bool = False,
            exact_match: bool = False
    ) -> "Event":
        """Get an :class:`Event` by its name.

        A fuzzy match is performed to find the event that best matches the
        given name. Fuzzy matching is performed using the country, location,
        name and officialName of each event. This is not guaranteed to return
        the correct result. You should therefore always check if the function
        actually returns the event you had wanted.
        To guarantee the function returns the event queried, toggle
        strict_search, which will only return an event if its event name
        matches (non case sensitive) the query string.

        .. warning:: You should avoid adding common words to ``name`` to avoid
            false string matches.
            For example, you should rather use "Belgium" instead of "Belgian
            Grand Prix" as ``name``.

        Args:
            name: The name of the event. For example,
                ``.get_event_by_name("british")`` and
                ``.get_event_by_name("silverstone")`` will both return the
                event for the British Grand Prix.
            strict_search: This argument is deprecated and planned for removal.
                Use the equivalent ``exact_match`` instead
            exact_match: Search only for exact query matches
                instead of using fuzzy search. For example,
                ``.get_event_by_name("British Grand Prix",
                exact_match=True)``
                will return the event for the British Grand Prix, whereas
                ``.get_event_by_name("British", exact_match=True)``
                will return ``None``
        """
        if strict_search:
            warnings.warn(("strict_search is deprecated and planned for "
                           "removal, use the equivalent exact_match instead"),
                          FutureWarning)
        if strict_search or exact_match:
            return self._strict_event_search(name)
        else:
            return self._fuzzy_event_search(name)


class Event(BaseSeries):
    """This class represents a single event (race weekend or testing event).

    Each event consists of one or multiple sessions, depending on the type
    of event and depending on the event format.

    For detailed information about the information that is available for each
    event, see `Event Schedule Data`_.

    This class is usually not instantiated directly. You should use
    :func:`fastf1.get_event` or similar to get a specific event.

    Args:
          year: Championship year
    """
    _metadata = ['year']

    def __init__(self, *args, year: int = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.year = year

    def is_testing(self) -> bool:
        """Return `True` or `False`, depending on whether this event is a
        testing event."""
        return self['EventFormat'] == 'testing'

    def get_session_name(self, identifier) -> str:
        """Return the full session name of a specific session from this event.

        Examples:

            >>> import fastf1
            >>> event = fastf1.get_event(2021, 1)
            >>> event.get_session_name(3)
            'Practice 3'
            >>> event.get_session_name('Q')
            'Qualifying'
            >>> event.get_session_name('praCtice 1')
            'Practice 1'

        Args:
            identifier: see :ref:`SessionIdentifier`

        Raises:
            ValueError: No matching session or invalid identifier
        """
        try:
            num = float(identifier)
        except ValueError:
            # by name or abbreviation
            for name in _SESSION_TYPE_ABBREVIATIONS.values():
                if identifier.casefold() == name.casefold():
                    session_name = name
                    break
            else:
                try:
                    session_name = \
                        _SESSION_TYPE_ABBREVIATIONS[identifier.upper()]
                except KeyError:
                    raise ValueError(f"Invalid session type '{identifier}'")

            # 'Sprint' was originally called 'Sprint Qualifying' only in the
            # old 'sprint' event format and renamed later; support the old
            # name for backwards compatibility by silently correcting to the
            # new name
            if ((session_name == 'Sprint Qualifying')
                    and (self.year in (2021, 2022))):
                session_name = 'Sprint'

            if session_name not in self.values:
                raise ValueError(f"Session type '{identifier}' does not "
                                 f"exist for this event")
        else:
            # by number
            if (float(num).is_integer()
                    and (num := int(num)) in (1, 2, 3, 4, 5)):
                session_name = self[f'Session{num}']
            else:
                raise ValueError(f"Invalid session type '{num}'")
            if not session_name:
                raise ValueError(f"Session number {num} does not "
                                 f"exist for this event")

        return session_name

    def get_session_date(self, identifier: Union[str, int], utc=False) \
            -> pd.Timestamp:
        """Return the date and time (if available) at which a specific session
        of this event is or was held.

        Args:
            identifier: see :ref:`SessionIdentifier`
            utc: return a non-timezone-aware UTC timestamp

        Raises:
            ValueError: No matching session or invalid identifier
        """
        session_name = self.get_session_name(identifier)
        relevant_columns = self.loc[['Session1', 'Session2', 'Session3',
                                     'Session4', 'Session5']]
        mask = (relevant_columns == session_name)
        if not mask.any():
            raise ValueError(f"Session type '{identifier}' does not exist "
                             f"for this event")
        else:
            _name = mask.idxmax()
            date_utc = self[f"{_name}DateUtc"]
            date = self[f"{_name}Date"]
            if (not utc) and pd.isnull(date) and (not pd.isnull(date_utc)):
                raise ValueError("Local timestamp is not available")
            if utc:
                return date_utc
            return date

    def get_session(self, identifier: Union[int, str]) -> "Session":
        """Return a session from this event.

        Args:
            identifier: see :ref:`SessionIdentifier`

        Raises:
            ValueError: No matching session or invalid identifier
        """
        try:
            num = float(identifier)
        except ValueError:
            # by name or abbreviation
            session_name = self.get_session_name(identifier)
            if session_name not in self.values:
                raise ValueError(f"Session type '{identifier}' does not "
                                 f"exist for this event")
        else:
            # by number
            if (float(num).is_integer()
                    and (num := int(num)) in (1, 2, 3, 4, 5)):
                session_name = self[f'Session{num}']
            else:
                raise ValueError(f"Invalid session type '{num}'")
            if not session_name:
                raise ValueError(f"Session number {num} does not "
                                 f"exist for this event")

        return Session(event=self, session_name=session_name,
                       f1_api_support=self.F1ApiSupport)

    def get_race(self) -> "Session":
        """Return the race session."""
        return self.get_session('Race')

    def get_qualifying(self) -> "Session":
        """Return the qualifying session."""
        return self.get_session('Qualifying')

    def get_sprint(self) -> "Session":
        """Return the sprint session."""
        return self.get_session('Sprint')

    def get_sprint_shootout(self) -> "Session":
        """Return the sprint shootout session."""
        return self.get_session('Sprint Shootout')

    def get_sprint_qualifying(self) -> "Session":
        """Return the sprint qualifying session."""
        return self.get_session('Sprint Qualifying')

    def get_practice(self, number: int) -> "Session":
        """Return the specified practice session.
        Args:
            number: 1, 2 or 3 - Free practice session number
        """
        return self.get_session(f'Practice {number}')



import warnings
from typing import (
    List,
    Optional
)

import numpy as np
import pandas as pd


try:
    import matplotlib
    from matplotlib import cycler
    from matplotlib import pyplot as plt
except ImportError:
    warnings.warn("Failed to import optional dependency 'matplotlib'!"
                  "Plotting functionality will be unavailable!",
                  RuntimeWarning)
try:
    import timple
except ImportError:
    warnings.warn("Failed to import optional dependency 'timple'!"
                  "Plotting of timedelta values will be restricted!",
                  RuntimeWarning)

from rapidfuzz import fuzz

from fastf1.logger import get_logger
from fastf1.plotting._constants import (
    LEGACY_DRIVER_COLORS,
    LEGACY_DRIVER_TRANSLATE,
    LEGACY_TEAM_COLORS,
    LEGACY_TEAM_TRANSLATE
)


_logger = get_logger(__package__)


_COLOR_PALETTE: List[str] = ['#FF79C6', '#50FA7B', '#8BE9FD', '#BD93F9',
                             '#FFB86C', '#FF5555', '#F1FA8C']
# The default color palette for matplotlib plot lines in fastf1's color scheme


class __DefaultStringArgType(str):
    pass


__color_scheme_default_arg = __DefaultStringArgType('fastf1')


def setup_mpl(
        mpl_timedelta_support: bool = True,
        color_scheme: Optional[str] = __color_scheme_default_arg,
        misc_mpl_mods: bool = True):
    """Setup matplotlib for use with fastf1.

    This is optional but, at least partly, highly recommended.

    .. deprecated:: 3.4.0

        The optional argument ``misc_mpls_mods`` is deprecated.

    .. deprecated:: 3.4.0

        The default value for ``color_scheme`` will change from ``'fastf1'``
        to ``None``. You should explicitly set the desired value when calling
        this function.


    Parameters:
        mpl_timedelta_support (bool):
            Matplotlib itself offers very limited functionality for plotting
            timedelta values. (Lap times, sector times and other kinds of time
            spans are represented as timedelta.)

            Enabling this option will patch some internal matplotlib functions
            and register converters, formatters and locators for tick
            formatting. The heavy lifting for this is done by an external
            package called 'Timple'. See https://github.com/theOehrly/Timple if
            you wish to customize the tick formatting for timedelta.

        color_scheme (str, None):
            This enables the Fast-F1 color scheme that you can see in all
            example images.
            Valid color scheme names are: ['fastf1', None]

        misc_mpl_mods (bool):
            This argument is deprecated since v3.4.0 and should no longer be
            used.
    """
    if color_scheme is __color_scheme_default_arg:
        warnings.warn(
            "FastF1 will no longer silently modify the default Matplotlib "
            "colors in the future.\nTo remove this warning, explicitly set "
            "`color_scheme=None` or `color_scheme='fastf1'` when calling "
            "`.setup_mpl()`.", FutureWarning
        )

    if misc_mpl_mods:
        warnings.warn(
            "FastF1 will stop modifying the default Matplotlib settings in "
            "the future.\nTo opt-in to the new behaviour and remove this "
            "warning, explicitly set `misc_mpl_mods=False` when calling "
            "`.setup_mpl()`.", FutureWarning
        )

    if mpl_timedelta_support:
        _enable_timple()
    if color_scheme == 'fastf1':
        _enable_fastf1_color_scheme()
    if misc_mpl_mods:
        _enable_misc_mpl_mods()


def driver_color(identifier: str) -> str:
    """
    Get a driver's color from a driver name or abbreviation.

    .. deprecated:: 3.4.0
        This function is deprecated and will be removed in a future version.
        Use :func:`~fastf1.plotting.get_driver_color` instead.

    This function will try to find a matching driver for any identifier string
    that is passed to it. This involves case-insensitive matching and partial
    string matching.

    Example::

        >>> driver_color('charles leclerc')  # doctest: +SKIP
        '#dc0000'
        >>> driver_color('max verstappen')  # doctest: +SKIP
        '#fcd700'
        >>> driver_color('ver')  # doctest: +SKIP
        '#fcd700'
        >>> driver_color('lec')  # doctest: +SKIP
        '#dc0000'

        shortened driver names and typos can be dealt with
        too (within reason)

        >>> driver_color('Max Verst')  # doctest: +SKIP
        '#fcd700'
        >>> driver_color('Charles')  # doctest: +SKIP
        '#dc0000'

    Args:
        identifier (str): Abbreviation or uniquely identifying name of the
            driver.

    Returns:
        str: hex color code
    """
    warnings.warn("The function `driver_color` is deprecated and will be "
                  "removed in a future version. Use "
                  "`fastf1.plotting.get_driver_color` instead.",
                  FutureWarning)

    if identifier.upper() in LEGACY_DRIVER_TRANSLATE:
        # try short team abbreviations first
        return LEGACY_DRIVER_COLORS[
            LEGACY_DRIVER_TRANSLATE[identifier.upper()]
        ]
    else:
        identifier = identifier.lower()

        # check for an exact team name match
        if identifier in LEGACY_DRIVER_COLORS:
            return LEGACY_DRIVER_COLORS[identifier]

        # check for exact partial string match
        for team_name, color in LEGACY_DRIVER_COLORS.items():
            if identifier in team_name:
                return color

        # do fuzzy string matching
        key_ratios = list()
        for existing_key in LEGACY_DRIVER_COLORS:
            ratio = fuzz.ratio(identifier, existing_key)
            key_ratios.append((ratio, existing_key))
        key_ratios.sort(reverse=True)
        if key_ratios[0][0] != 100:
            _logger.warning(
                ("Correcting invalid user input "
                 f"'{identifier}' to '{key_ratios[0][1]}'."
                 )
            )
        if ((key_ratios[0][0] < 35)
                or (key_ratios[0][0] / key_ratios[1][0] < 1.2)):
            # ensure that the best match has a minimum accuracy (35 out of
            # 100) and that it has a minimum confidence (at least 20% better
            # than second best)
            raise KeyError
        best_matched_key = key_ratios[0][1]
        return LEGACY_DRIVER_COLORS[best_matched_key]


def team_color(identifier: str) -> str:
    """
    Get a team's color from a team name or abbreviation.

    .. deprecated:: 3.4.0
        This function is deprecated and will be removed in a future version.
        Use :func:`~fastf1.plotting.get_team_color` instead.

    This function will try to find a matching team for any identifier string
    that is passed to it. This involves case-insensitive matching and partial
    string matching.

    Example::

        >>> team_color('Red Bull')  # doctest: +SKIP
        '#fcd700'
        >>> team_color('redbull')  # doctest: +SKIP
        '#fcd700'
        >>> team_color('Red')  # doctest: +SKIP
        '#fcd700'
        >>> team_color('RBR')  # doctest: +SKIP
        '#fcd700'

        # shortened team names, included sponsors and typos can be dealt with
        # too (within reason)

        >>> team_color('Mercedes')  # doctest: +SKIP
        '#00d2be'
        >>> team_color('Merc')  # doctest: +SKIP
        '#00d2be'
        >>> team_color('Merecds')  # doctest: +SKIP
        '#00d2be'
        >>> team_color('Mercedes-AMG Petronas F1 Team')  # doctest: +SKIP
        '#00d2be'

    Args:
        identifier (str): Abbreviation or uniquely identifying name of the
            team.

    Returns:
        str: hex color code
    """
    warnings.warn("The function `team_color` is deprecated and will be "
                  "removed in a future version. Use "
                  "`fastf1.plotting.get_team_color` instead.",
                  FutureWarning)

    if identifier.upper() in LEGACY_TEAM_TRANSLATE:
        # try short team abbreviations first
        return LEGACY_TEAM_COLORS[LEGACY_TEAM_TRANSLATE[identifier.upper()]]
    else:
        identifier = identifier.lower()
        # remove common non-unique words
        for word in ('racing', 'team', 'f1', 'scuderia'):
            identifier = identifier.replace(word, "")

        # check for an exact team name match
        if identifier in LEGACY_TEAM_COLORS:
            return LEGACY_TEAM_COLORS[identifier]

        # check for exact partial string match
        for team_name, color in LEGACY_TEAM_COLORS.items():
            if identifier in team_name:
                return color

        # do fuzzy string matching
        key_ratios = list()
        for existing_key in LEGACY_TEAM_COLORS.keys():
            ratio = fuzz.ratio(identifier, existing_key)
            key_ratios.append((ratio, existing_key))
        key_ratios.sort(reverse=True)
        if key_ratios[0][0] != 100:
            _logger.warning(
                ("Correcting invalid user input "
                 f"'{identifier}' to '{key_ratios[0][1]}'."
                 )
            )
        if ((key_ratios[0][0] < 35)
                or (key_ratios[0][0] / key_ratios[1][0] < 1.2)):
            # ensure that the best match has a minimum accuracy (35 out of
            # 100) and that it has a minimum confidence (at least 20% better
            # than second best)
            raise KeyError
        best_matched_key = key_ratios[0][1]
        return LEGACY_TEAM_COLORS[best_matched_key]


def _enable_timple():
    # use external package timple to patch matplotlib
    # this adds converters, locators and formatters for
    # plotting timedelta values
    tick_formats = [
        "%d %day",
        "%H:00",
        "%H:%m",
        "%M:%s.0",
        "%M:%s.%ms"
    ]
    tmpl = timple.Timple(converter='concise',
                         formatter_args={'show_offset_zero': False,
                                         'formats': tick_formats})
    tmpl.enable()


def _enable_misc_mpl_mods():
    def _bar_sorted(bar):
        def _bar_sorted_decorator(*args, **kwargs):
            if 'edgecolor' not in kwargs:
                kwargs['edgecolor'] = 'none'
            if 'sort' in kwargs and len(val := args[-1]):
                s = kwargs['sort']
                if s == 'increasing' or s == 1:
                    s = False
                if s == 'decreasing' or s == -1:
                    s = True
                _ids = [list(val).index(a) for a in sorted(val, reverse=s)]
                _args = [[args[-2][i] for i in _ids],
                         [args[-1][i] for i in _ids]]
                if len(args) > 2:
                    _args.insert(0, args[0])
                args = _args
                for key in kwargs:
                    if isinstance(kwargs[key], (pd.core.series.Series)):
                        kwargs[key] = kwargs[key].to_numpy()
                    if isinstance(kwargs[key], (list, np.ndarray)):
                        kwargs[key] = [kwargs[key][i] for i in _ids]
                kwargs.pop('sort', None)
            return bar(*args, **kwargs)

        return _bar_sorted_decorator

    plt.bar = _bar_sorted(plt.bar)
    plt.barh = _bar_sorted(plt.barh)
    matplotlib.axes.Axes.bar = _bar_sorted(matplotlib.axes.Axes.bar)
    matplotlib.axes.Axes.barh = _bar_sorted(matplotlib.axes.Axes.barh)

    def _nice_grid(ax):
        if isinstance(ax, np.ndarray):
            [_nice_grid(_ax) for _ax in ax]
        else:
            ax.minorticks_on()
            grid = getattr(ax, 'grid')
            grid(visible=True, which='major', color='#4f4845',
                 linestyle='-', linewidth=1)
            grid(visible=True, which='minor', color='#3f3a38',
                 linestyle='--', linewidth=0.5)

    _subplots_placeholder = plt.subplots

    def _subplots(*args, **kwargs):
        fig, ax = _subplots_placeholder(*args, **kwargs)
        _nice_grid(ax)
        return fig, ax

    plt.subplots = _subplots

    _savefig_placeholder = matplotlib.figure.Figure.savefig

    def _save(*args, **kwargs):
        if 'facecolor' not in kwargs:
            kwargs['facecolor'] = args[0].get_facecolor()
        if 'edgecolors' not in kwargs:
            kwargs['edgecolor'] = 'none'
        return _savefig_placeholder(*args, **kwargs)

    matplotlib.figure.Figure.savefig = _save


def _enable_fastf1_color_scheme():
    plt.rcParams['figure.facecolor'] = '#292625'
    plt.rcParams['axes.edgecolor'] = '#2d2928'
    plt.rcParams['xtick.color'] = '#f1f2f3'
    plt.rcParams['ytick.color'] = '#f1f2f3'
    plt.rcParams['axes.labelcolor'] = '#F1f2f3'
    plt.rcParams['axes.facecolor'] = '#1e1c1b'
    # plt.rcParams['axes.facecolor'] = '#292625'
    plt.rcParams['axes.titlesize'] = 'x-large'
    # plt.rcParams['font.family'] = 'Gravity'
    plt.rcParams['font.weight'] = 'medium'
    plt.rcParams['text.color'] = '#F1F1F3'
    plt.rcParams['axes.titlesize'] = '19'
    plt.rcParams['axes.titlepad'] = '12'
    plt.rcParams['axes.titleweight'] = 'light'
    plt.rcParams['axes.prop_cycle'] = cycler('color', _COLOR_PALETTE)
    plt.rcParams['legend.fancybox'] = False
    plt.rcParams['legend.facecolor'] = (0.1, 0.1, 0.1, 0.7)
    plt.rcParams['legend.edgecolor'] = (0.1, 0.1, 0.1, 0.9)
    plt.rcParams['savefig.transparent'] = False
    plt.rcParams['axes.axisbelow'] = True


def lapnumber_axis(ax, axis='xaxis'):
    """
    Set axis to integer ticks only.

    .. deprecated:: 3.4.0
        The function ``lapnumber_axis`` is deprecated and will ber removed in a
        future version without replacement.

    Args:
        ax: matplotlib axis
        axis: can be 'xaxis' or 'yaxis'

    Returns:
        the modified axis instance
    """
    warnings.warn("The function `lapnumber_axis` is deprecated and will be "
                  "removed without replacement in a future version.",
                  FutureWarning)
    getattr(ax, axis).get_major_locator().set_params(integer=True,
                                                     min_n_ticks=0)

    return ax



"""
.. _logging:

Logging - :mod:`fastf1.logger`
==============================

FastF1 uses loggers for each submodule. All loggers are child loggers of
FastF1's base logger. The log level is usually configured equally for all parts
of FastF1. The :class:`~fastf1.logger.LoggingManager` or the direct access
functions should commonly be used for this.

Note to developers: some parts of FastF1's data loading are wrapped in generic
catch-all error handling to prevent errors in individual data loading tasks to
make FastF1 completely unusable. Instead, unhandled exceptions will be caught,
short error message is logged in level INFO, the full traceback is logged on
level DEBUG and execution will continue as good as possible. This system can
make debugging more difficult because errors are not raised. To circumvent
this, there are two possible ways to disable the catch-all error handling for
data loading:

- explicitly set :attr:`fastf1.logger.LoggingManager.debug` to `True`
- set the environment variable `FASTF1_DEBUG=1`


Logging Manager
...............

.. autoclass:: fastf1.logger.LoggingManager
    :members:


Functions for direct access
...........................

.. automethod:: fastf1.logger.set_log_level
.. automethod:: fastf1.logger.get_logger
.. automethod:: fastf1.logger.soft_exceptions

"""

import functools
import logging
import os
import warnings
from typing import Union


class LoggingManager:
    """Interface for configuring logging in FastF1.

    All parts of FastF1 generally log at the log level 'INFO'.
    The reason for this is that many data loading processes take multiple
    seconds to complete. Logging is used to give progress information as well
    as for showing warnings and non-terminal errors.

    All submodule loggers in FastF1 are child loggers of the base logger.
    This class acts as an interface to set the log level for FastF1 and get
    child loggers.
    """
    _console_formatter = logging.Formatter(
        "{module: <8} {levelname: >10} \t{message}", style='{'
    )

    _console_handler = logging.StreamHandler()
    _console_handler.setFormatter(_console_formatter)
    _console_handler.setLevel(logging.INFO)

    _root_logger = logging.getLogger('fastf1')
    _root_logger.setLevel(logging.DEBUG)
    _root_logger.addHandler(_console_handler)

    debug = False
    """Flag for enabling debug mode. This will disable catch-all error handling
    for data loading methods."""

    @classmethod
    def get_child(cls, name: str):
        """Return a logger with the given name that is child of the base
        logger.

        Args:
            name: name of the child logger
        """
        return cls._root_logger.getChild(name)

    @classmethod
    def set_level(cls, level: int):
        """Set the log level for FastF1.

        Args:
            level: log level, for example `logging.INFO`
        """
        cls._console_handler.setLevel(level)


if os.getenv('FASTF1_DEBUG') == '1':
    warnings.warn("Debug Mode enabled for Logger!", UserWarning)
    LoggingManager.debug = True
else:
    LoggingManager.debug = False


def get_logger(name: str):
    """Return a logger with the given name that is a child of FastF1's
    base logger.
    """
    return LoggingManager.get_child(name)


def set_log_level(level: Union[str, int]):
    """Set the log level for all parts of FastF1.

    When setting the log level for FastF1, only messages with this level or
    with a higher level will be shown.

    Args:
        level: Either a log level from the logging module (e.g. `logging.INFO`)
            or the level as a string (e.g. 'WARNING').
    """
    if isinstance(level, str):
        level = logging._nameToLevel.get(level.upper())
    LoggingManager.set_level(level)


def soft_exceptions(descr_name: str, msg: str, logger: logging.Logger):
    """Wrapper method for wrapping any function into catch-all error handling
    that can be disabled by setting :attr:`~fastf1.logger.LoggingManager.debug`
    to `True`.

    Args:
        descr_name: descriptive name for the type of data that should have
            been loaded by the wrapped function
        msg: Short message that is shown as error message to users
        logger: the logger that should be used to log errors (a logger instance
            as returned by :func:`get_logger`, for example).
    """
    # This function is used to wrap individual data loading functions that are
    # called by `Session.load`. With the default configuration, this wrapper
    # will catch all unhandled exceptions in the wrapped function and log them.
    # The idea is that in case of an error, data loading will only fail
    # partially and FastF1 will not become completely unusable.
    # For development purposes the automatic error handling can be disabled
    # by explicitly setting `Logger.debug = True` or by setting the environment
    # variable `FASTF1_DEBUG=1`. In this case, all unhandled exceptions will
    # be raised.
    def __decorator(func):
        @functools.wraps(func)
        def __wrapped(*args, **kwargs):
            if not LoggingManager.debug:
                try:
                    return func(*args, **kwargs)
                except Exception as exc:
                    logger.warning(msg)
                    logger.debug(f"Traceback for failure in {descr_name}",
                                 exc_info=exc)
            else:
                return func(*args, **kwargs)

        return __wrapped
    return __decorator



from dataclasses import dataclass
from typing import Dict


class Compounds:
    HyperSoft = "HYPERSOFT"
    UltraSoft = "ULTRASOFT"
    SuperSoft = "SUPERSOFT"
    Soft = "SOFT"
    Medium = "MEDIUM"
    Hard = "HARD"
    SuperHard = "SUPERHARD"
    Intermediate = "INTERMEDIATE"
    Wet = "WET"
    Unknown = "UNKNOWN"
    TestUnknown = "TEST-UNKNOWN"


@dataclass(frozen=True)
class TeamColors:
    Official: str
    Default: str


@dataclass(frozen=True)
class Team:
    ShortName: str
    TeamColor: TeamColors


@dataclass(frozen=True)
class BaseSeason:
    CompoundColors: Dict[str, str]
    Teams: Dict[str, Team]



import warnings
from typing import (
    Dict,
    List
)

from fastf1.plotting._constants import \
    LEGACY_DRIVER_COLORS as _LEGACY_DRIVER_COLORS
from fastf1.plotting._constants import \
    LEGACY_DRIVER_TRANSLATE as _LEGACY_DRIVER_TRANSLATE
from fastf1.plotting._constants import \
    LEGACY_TEAM_TRANSLATE as _LEGACY_TEAM_TRANSLATE
from fastf1.plotting._constants import Constants as _Constants
from fastf1.plotting._interface import (  # noqa: F401
    _get_driver_team_mapping,
    add_sorted_driver_legend,
    get_compound_color,
    get_compound_mapping,
    get_driver_abbreviation,
    get_driver_abbreviations_by_team,
    get_driver_color,
    get_driver_color_mapping,
    get_driver_name,
    get_driver_names_by_team,
    get_driver_style,
    get_team_color,
    get_team_name,
    get_team_name_by_driver,
    list_compounds,
    list_driver_abbreviations,
    list_driver_names,
    list_short_team_names,
    list_team_names
)
from fastf1.plotting._plotting import (  # noqa: F401
    _COLOR_PALETTE,
    driver_color,
    lapnumber_axis,
    setup_mpl,
    team_color
)


__all__ = [
    # imported, current
    'add_sorted_driver_legend',
    'get_compound_color',
    'get_compound_mapping',
    'get_driver_abbreviation',
    'get_driver_abbreviations_by_team',
    'get_driver_color',
    'get_driver_color_mapping',
    'get_driver_name',
    'get_driver_names_by_team',
    'get_driver_style',
    'get_team_color',
    'get_team_name',
    'get_team_name_by_driver',
    'list_compounds',
    'list_driver_abbreviations',
    'list_driver_names',
    'list_team_names',
    'list_short_team_names',
    'setup_mpl',

    # imported, legacy
    'driver_color',
    'lapnumber_axis',
    'team_color',

    # legacy
    'COMPOUND_COLORS',
    'DRIVER_COLORS',
    'DRIVER_TRANSLATE',
    'TEAM_COLORS',
    'TEAM_TRANSLATE',
    'COLOR_PALETTE'
]


def __getattr__(name):
    if name in ('COMPOUND_COLORS', 'DRIVER_TRANSLATE', 'DRIVER_COLORS',
                'TEAM_COLORS', 'TEAM_TRANSLATE', 'COLOR_PALETTE'):
        warnings.warn(f"{name} is deprecated and will be removed in a future "
                      f"version.", FutureWarning)

        return globals()[f"_DEPR_{name}"]

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


_DEPR_COMPOUND_COLORS: Dict[str, str] = {
    key: val for key, val
    in _Constants['2024'].CompoundColors.items()
}
COMPOUND_COLORS: Dict[str, str]
"""
Mapping of tyre compound names to compound colors (hex color codes).
(current season only)

.. deprecated:: 3.4.0
    The ``COMPOUND_COLORS`` dictionary is deprecated and will be removed in a
    future version. Use :func:`~fastf1.plotting.get_compound_color` or
    :func:`~fastf1.plotting.get_compound_mapping` instead.
"""


_DEPR_DRIVER_COLORS: Dict[str, str] = _LEGACY_DRIVER_COLORS.copy()
DRIVER_COLORS: Dict[str, str]
"""
Mapping of driver names to driver colors (hex color codes).

.. warning::
    This dictionary will no longer be updated to include new drivers. Use
    the new API instead.

.. deprecated:: 3.4.0
    The ``DRIVER_COLORS`` dictionary is deprecated and will ber removed in a
    future version. Use :func:`~fastf1.plotting.get_driver_color` or
    :func:`~fastf1.plotting.get_driver_color_mapping` instead.
"""


_DEPR_DRIVER_TRANSLATE: Dict[str, str] = _LEGACY_DRIVER_TRANSLATE.copy()
DRIVER_TRANSLATE: Dict[str, str]
"""
Mapping of driver names to theirs respective abbreviations.

.. warning::
    This dictionary will no longer be updated to include new drivers. Use
    the new API instead.


.. deprecated:: 3.4.0
    The ``DRIVER_TRANSLATE`` dictionary is deprecated and will be removed in a
    future version. Use :func:`~fastf1.plotting.get_driver_name` instead.
"""

_DEPR_TEAM_COLORS: Dict[str, str] = {
    # str(key.value): val for key, val
    # in _Constants['2024'].Colormaps[_Colormaps.Default].items()
    name: team.TeamColor.Default for name, team
    in _Constants['2024'].Teams.items()
}
TEAM_COLORS: Dict[str, str]
"""
Mapping of team names to team colors (hex color codes).
(current season only)

.. deprecated:: 3.4.0
    The ``TEAM_COLORS`` dictionary is deprecated and will be removed in a
    future version. Use :func:`~fastf1.plotting.get_team_color` instead.
"""

_DEPR_TEAM_TRANSLATE: Dict[str, str] = _LEGACY_TEAM_TRANSLATE.copy()
TEAM_TRANSLATE: Dict[str, str]
"""
Mapping of team names to theirs respective abbreviations.

.. deprecated:: 3.4.0
    The ``TEAM_TRANSLATE`` dictionary is deprecated and will be removed in a
    future version. Use :func:`~fastf1.plotting.get_team_name` instead.
"""

_DEPR_COLOR_PALETTE: List[str] = _COLOR_PALETTE.copy()
COLOR_PALETTE: List[str]
"""
The default color palette for matplotlib plot lines in fastf1's color scheme.

.. deprecated:: 3.4.0
    The ``COLOR_PALETTE`` list is deprecated and will be removed in a
    future version with no replacement.
"""



import warnings
from typing import (
    Dict,
    List
)

import fastf1._api
from fastf1.plotting._base import (
    _Driver,
    _normalize_string,
    _Team
)
from fastf1.plotting._constants import Constants


def _load_drivers_from_f1_livetiming(
        *, api_path: str, year: str
) -> List[_Team]:
    # load the driver information for the determined session
    driver_info = fastf1._api.driver_info(api_path)

    # parse the data into the required format
    teams: Dict[str, _Team] = dict()

    # Sorting by driver number here will directly guarantee that drivers
    # are sorted by driver number within each team. This has two advantages:
    # - the driver index in a team is consistent as long as the drivers don't
    #   change/reserver drivers are used/...
    # - the reigning champion (number 1) always has index 0, i.e. gets the
    #   primary style
    for num in sorted(driver_info.keys()):
        driver_entry = driver_info[num]
        team_name = driver_entry.get('TeamName')

        if team_name in teams:
            team = teams[team_name]
        else:
            team = _Team()
            team.value = team_name

        abbreviation = driver_entry.get('Tla')

        name = ' '.join((driver_entry.get('FirstName'),
                         driver_entry.get('LastName')))
        driver = _Driver()
        driver.value = name
        driver.normalized_value = _normalize_string(name).lower()
        driver.abbreviation = abbreviation
        driver.team = team

        team.drivers.append(driver)

        if team not in teams:
            normalized_full_team_name = _normalize_string(team_name).lower()
            for ref_team_name in Constants[year].Teams.keys():
                if ref_team_name in normalized_full_team_name:
                    team.normalized_value = ref_team_name
                    break
            else:
                warnings.warn(f"Encountered unknown team '{team_name}' while "
                              f"loading driver-team mapping.",
                              UserWarning)
                continue

            teams[team_name] = team

    return list(teams.values())
